/*
    dtaus.c - Belegloser Datenträgeraustausch mit einer Bank
    Copyright (c) 1996,8,2001-4,5,9  Martin Schulze <joey@infodrom.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111, USA.

    $Id: dtaus.c,v 1.48 2009-07-26 18:05:51 joey Exp $
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <malloc.h>

#include "dtaus.h"
#include "sepadata.h"

#include "memory.h"

static char *   s_szReadBuffer	= NULL;
static int      s_iBufferLen	= 0;


#define UmlauteAusschreiben

/*
 *  First: Some control structures
 */

typedef struct
{
  char *		name;
  unsigned short int	len;
  int             	type;

} dtaus_record;

#define REC_A	1
#define REC_C	2
#define REC_E	3

#define REQ	1
#define OPT	2
#define IGN	3

dtaus_record    recA[] = {
  {"Art", 		10, REQ},
  {"Name", 		27, REQ},
  {"IBAN", 		22, REQ},
  {"BIC", 		11, REQ},
  {"Referenz", 		18, OPT},
  {"Datum", 		 6, IGN},
  {"Ausfuehrung",	 8, OPT},
  {"BatchBooking", 	 0, OPT},

  {NULL, 0, 0}
};



enum A_RecordID
{
  A_TRANS	 = 0,
  A_NAME,	/* 1 */
  A_IBAN,	/* 2 */
  A_BIC,	/* 3 */
  A_REF,	/* 4 */
  A_DATE,	/* 5 */
  A_WHEN,	/* 6 */
  A_BBooking,	/* 7 */

  A_LEN
};



dtaus_record    recC[] = {
  {"Mandat", 		 70, OPT},
  {"SigDatum", 		 10, OPT},
  {"Name", 		 70, REQ},
  {"IBAN", 		 22, OPT},
  {"BIC", 		 11, OPT},
  {"Transaktion", 	 25, REQ},
  {"Betrag", 		 11, REQ},
  {"Zweck", 		140, REQ},
  {"Text", 		140, OPT},

  {NULL, 0, 0}
};

enum C_RecordID
{
  C_MANDATE	 = 0,
  C_MNDTDOS,	/* 1 DateOfSign*/
  C_NAME,	/* 2 */
  C_IBAN,	/* 3 */
  C_BIC,	/* 4 */
  C_TRANS,	/* 5 */
  C_VAL,	/* 6 */
  C_PURPOSE,	/* 7 */
  C_TEXT,	/* 8 */
  C_EXT, 	/* 9 */

  C_LEN
};

#define MAX_TEXT 14



  /*
   * [A-Z]:	 		0x01
   * [0-9]:	 		0x02
   * [A-Z2-9]:	 		0x04
   * [A-NP-Z0-9]:		0x08
   * [A-Z]:	 		0x10
   * [a-zA-Z0-9':?, (+,)/]:	0x80
   */

static unsigned char szCharClass[128] = 
  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x80, 0x80, 0x00, 0x80, 0x80, 0x80, 0x80, 0x80,
    0x8A, 0x8A, 0x8E, 0x8E, 0x8E, 0x8E, 0x8E, 0x8E, 0x8E, 0x8E, 0x80, 0x00, 0x00, 0x00, 0x00, 0x80,
    0x00, 0x8D, 0x8D, 0x8D, 0x8D, 0x8D, 0x8D, 0x8D, 0x8D, 0x8D, 0x8D, 0x8D, 0x8D, 0x8D, 0x8D, 0x85,
    0x8D, 0x8D, 0x8D, 0x8D, 0x8D, 0x8D, 0x8D, 0x8D, 0x8D, 0x8D, 0x8D, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90,
    0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x00, 0x00, 0x00, 0x00, 0x00};




/***********************************************************************************
 * prototypes
 ***********************************************************************************/
static int			_iGetCharLength (int iByte);
static int			_iGetUnicodeValue (unsigned char * szChar, int iMaxLength, int * piCharLength);
static char *	           	downcase (char *s);
static char *     		extract_ident (char *line);
static int			rec_index (char *ident, int type);
static size_t			control_nextline (char * szOutBuf, int iOutBufLen, FILE * ptInFile);
static char *			get_date ();
static int			dtaus_writeA (struct t_Globals * ptGlobals, char **values);
static int			dtaus_writeC (struct t_Globals * ptGlobals, char **valuesA, char **values, char ** aszText);
static int			_GetTransactionSubType(struct t_Globals * ptGlobals, char * szType);
static struct  t_CredDebt *	_GetCredDebt(char ** pszName, char ** pszIBAN, char ** pszBIC);
static int 			_CheckBIC(char * szBIC);
static int			_CheckIBAN(char * szIBAN);
static int			_GetValueInCent(char * szValue);
static char *			_GetPurposeString(char * szPurpose, int iPurposeLen, int iAddTextLines, char ** aszText);
static char *			_GetISOString(char * szSource, char * szDest, int iMaxDestLen, int iUpperCase);
static time_t			_GetISODate(char * szDate, int iCheckPast);
static int			_GiveUpTransactions(struct t_Globals * ptGlobals);
static int			_GiveUpInitiator(struct t_Globals * ptGlobals);
static int			_GetLastInserted(struct t_Globals * ptGlobals, int * iValueInCent, struct t_CredDebt ** ptPayment);

/***********************************************************************************
 ***********************************************************************************/
int dtaus_writeA (struct t_Globals * ptGlobals, char **values)
{
  int             i;


  for (i = 0; (recA[i].name); i++)
    if ((recA[i].type == REQ) && !values[i])
    {
      fprintf (stderr, "error: header record not complete, %s missing\n",
	       recA[i].name);
      return 0;
    }

  i = A_NAME;
  if (values[i] && strlen (values[i]) > recA[i].len)
  {
    fprintf (stderr, "warning: header name exceed the max length of %d chars\n", recA[i].len);
    values[i][recA[i].len] = '\0';
  }

  if (strlen(values[A_IBAN]) != recA[A_IBAN].len)
  {
    fprintf (stderr, "error: header %s: invalid length\n", recA[A_IBAN].name);
    return (0);
  }
  if (strlen(values[A_BIC]) != recA[A_BIC].len)
  {
    fprintf (stderr, "error: header %s: invalid length\n", recA[A_BIC].name);
    return (0);
  }


  ptGlobals->ptInitiator = _GetCredDebt( & values[A_NAME], & values[A_IBAN], & values[A_BIC]);
  if ( ! ptGlobals->ptInitiator)
  {
    fprintf (stderr, "error: header: invalid data for initiator\n");
    return (0);
  }

  ptGlobals->szIdentification			= values[A_REF];

  {
    struct tm tTime;
    time_t    iTime = time(NULL);

    localtime_r( & iTime, & tTime);


    if (values[A_WHEN])
    {
      iTime = _GetISODate(values[A_WHEN], 1);

      if (iTime == 0)
      {
        fprintf (stderr, "error: invalid date: %s\n", values[A_WHEN]);
        return (0);
      }

    }
    else
    {

      tTime.tm_sec = 0;
      tTime.tm_min = 0;
      if (tTime.tm_hour > 16)
        tTime.tm_mday++;
      tTime.tm_hour = 0;

      iTime = mktime( & tTime);
    }

    ptGlobals->iExecTimeStamp = iTime;
  }
  



  if (strcmp(values[A_TRANS], "TRF") == 0)
  {
    ptGlobals->iAction = eAction_CreditTransfer;
    ptGlobals->ptDebitor = ptGlobals->ptInitiator;
  }
  else if (strcmp(values[A_TRANS], "DDCORE") == 0)
  {
    ptGlobals->iAction = eAction_DirectDebitCore;
    ptGlobals->ptCreditor = ptGlobals->ptInitiator;
  }
  else if (strcmp(values[A_TRANS], "DDB2B") == 0)
  {
    ptGlobals->iAction = eAction_DirectDebitB2B;
    ptGlobals->ptCreditor = ptGlobals->ptInitiator;
  }
  else
  {
    fprintf (stderr, "error: invalid transaction type: %s, possible types TRF, DDCORE, DDB2B\n", values[A_TRANS]);
    return (0);
  }

  return 1;
}

/***********************************************************************************
 *  valuesA		RecA values
 *  values		RecC values
 *  aszText		additional C_TEXT lines - only set if values[C_TEXT] is set 
 * return
 *  int			0: fault
 *			1: success
 *                     -1: record does not match output criteria
 ***********************************************************************************/
int dtaus_writeC (struct t_Globals * ptGlobals, char **valuesA, char **values, char ** aszText)
{
  int             iI;
  int             iAddTextLines;

  /* check if all mandatory entries are made */
  for (iI = 0; (recC[iI].name); iI++)
  {
    if ((recC[iI].type == REQ) && !values[iI])
    {
      fprintf (stderr, "error: transaction #%d: mandatory entry not present: %s\n", ptGlobals->iPmtCount + 1, recC[iI].name);
      return 0;
    }
    if (values[iI] && (strlen(values[iI]) > recC[iI].len))
    {
      fprintf (stderr, "error: transaction #%d: invalid string length: %s (%d allowed)\n", ptGlobals->iPmtCount + 1, recC[iI].name,  recC[iI].len);
      return (0);
    }
    
  }

  enum eTASubType 	iSubType	= _GetTransactionSubType(ptGlobals, values[C_TRANS]);
  if (iSubType == eTASubType_InvalidType)
  {
    fprintf (stderr, "error: transaction #%d: invalid %s\n", ptGlobals->iPmtCount + 1, recC[C_TRANS].name);
    return (0);
  }


  int			iValue		= _GetValueInCent(values[C_VAL]);
  if (iValue < 0)
  {
    fprintf (stderr, "error: transaction #%d: invalid %s\n", ptGlobals->iPmtCount + 1, recC[C_VAL].name);
    return (0);
  }

  aszText[0] = values[C_TEXT];
  values[C_TEXT] = NULL;  
  for (iAddTextLines = 0; aszText && aszText[iAddTextLines]; iAddTextLines++);
  char *		szPurpose	= _GetPurposeString(values[C_PURPOSE], recC[C_PURPOSE].len, iAddTextLines, aszText);
  if (szPurpose == NULL)
  {
    fprintf (stderr, "error: transaction #%d: invalid %s/%s\n", ptGlobals->iPmtCount + 1, recC[C_PURPOSE].name, recC[C_TEXT].name);
    return (0);
  }

  if (ptGlobals->iAction == eAction_CreditTransfer)
  {

    struct t_CredDebt *	ptCredDebt	= _GetCredDebt( & values[C_NAME], & values[C_IBAN], & values[C_BIC]);
    if (ptCredDebt == NULL)
    {
      fprintf (stderr, "error: transaction #%d: invalid creditor data\n", ptGlobals->iPmtCount + 1);
      return (0);
    }

    /* new list element */
    struct t_CdtTrfTx * ptNew = (struct t_CdtTrfTx *) malloc (sizeof(struct t_CdtTrfTx));

    /* fill the data */
    ptNew->iType	= iSubType;
    ptNew->ptCreditor	= ptCredDebt;
    ptNew->iValueInCent	= iValue;
    ptNew->szPurpose	= szPurpose;
    ptNew->ptNext	= NULL;

    /* adjust the global counters for the elements */
    ptGlobals->iPmtCount++;
    ptGlobals->iPmtSumInCent += iValue;

    /* chain new entry to the end of the list */
    struct t_CdtTrfTx ** ptRun = & ptGlobals->ptTxList.ptCdtTrfTxList;
    while (*ptRun != NULL)
      ptRun = & (*ptRun)->ptNext;

    *ptRun = ptNew;

    /* keep new element as last element in mind */
    ptGlobals->ptLICdtTrfTx = ptNew;

  } //////// *********************************************************
  ////////// ********************************************************* 
  ////////// ********************************************************* 
  ////////// ********************************************************* 
  else ///// direct debit transfer
  ////////// ********************************************************* 
  ////////// ********************************************************* 
  ////////// ********************************************************* 
  { //////// *********************************************************

    //**************************************+
    /* check mandatory entries */
    //**************************************+

    char * szMandate = _GetISOString(values[C_MANDATE], NULL, recC[C_VAL].len, 0);
    if (szMandate == NULL)
    {
      fprintf (stderr, "error: transaction #%d: invalid or missing mandate\n", ptGlobals->iRecordCount + 1);
      return (0);
    }

    char * szDebtName = _GetISOString(values[C_NAME], NULL, recC[C_NAME].len, 0);
    if (szMandate == NULL)
    {
      fprintf (stderr, "error: transaction #%d: invalid or missing name\n", ptGlobals->iRecordCount + 1);
      return (0);
    }

    //**************************************+
    /* check consistency if present */
    //**************************************+
    char * szIBAN = NULL;
    if (values[C_IBAN])
    {
      if ( ! _CheckIBAN(values[C_IBAN]))
      {
        fprintf (stderr, "error: transaction #%d: invalid entry %s\n", ptGlobals->iRecordCount + 1, recC[C_IBAN].name);
        return (0);
      }

      szIBAN = values[C_IBAN];
      values[C_IBAN] = NULL;
    }

    char * szBIC = NULL;
    if (values[C_BIC])
    {
      if ( ! _CheckBIC(values[C_BIC]))
      {
        fprintf (stderr, "error: transaction #%d: invalid entry %s\n", ptGlobals->iRecordCount + 1, recC[C_BIC].name);
        return (0);
      }

      szBIC = values[C_BIC];
      values[C_BIC] = NULL;
    }
    
    time_t lDateOfSign = 0;
    if (values[C_MNDTDOS])
    {
      lDateOfSign = _GetISODate(values[C_MNDTDOS], 0);
      if (lDateOfSign == 0)
      {
        fprintf (stderr, "error: transaction #%d: invalid entry %s\n", ptGlobals->iRecordCount + 1, recC[C_MNDTDOS].name);
        return (0);
      }
    }
    


    //********************************************+
    /* switch if it's the first usage of mandate */
    //********************************************+

    int			iCreateNew	= 0;
    enum eMndtChange	iChange		= eMndtChange_None;

    struct t_Mandate *	ptRootMandate	= FindMandate(ptGlobals, szMandate);
    if ( ! ptRootMandate)
    {
      //********************************************+
      /* first usage */
      //********************************************+
      if (( ! szIBAN) || ( ! szBIC) || ( ! lDateOfSign))
      {
        fprintf (stderr, "error: transaction #%d: mandatory entries %s, %s and/or %s is/are missing\n", ptGlobals->iRecordCount + 1, 
                         recC[C_IBAN].name, recC[C_BIC].name, recC[C_MNDTDOS].name);
        return (0);
      }
      iCreateNew = 1;

      if (iSubType == eTASubType_DirectDebitFinal)
      {
        fprintf (stderr, "warning: transaction #%d: invalid subtype final for new mandate using one-off\n", ptGlobals->iRecordCount + 1);
        iSubType = eTASubType_DirectDebitOneOff;
      }
      else if (iSubType == eTASubType_DirectDebitRepeat)
        // switching repeat to first
        iSubType = eTASubType_DirectDebitFirst;
    }
    else
    {
      //********************************************+
      /* repetitive usage */
      //********************************************+
      if (iSubType == eTASubType_DirectDebitOneOff)
      {
        fprintf (stderr, "error: transaction #%d: invalid subtype one off for an used mandate\n", ptGlobals->iRecordCount + 1);
        return (0);
      }
      else if (iSubType == eTASubType_DirectDebitFirst)
        iSubType = eTASubType_DirectDebitRepeat;


      if (szBIC && (strcmp(szBIC, ptRootMandate->ptNewestVersion->tDebitor.szBIC) != 0))
      {
        iCreateNew = 1;
        iChange = eMndtChange_BIC;

        if (iSubType == eTASubType_DirectDebitFinal)
        {
          fprintf (stderr, "warning: transaction #%d: invalid subtype final for a bank switching mandate using one-off\n", ptGlobals->iRecordCount + 1);
          iSubType = eTASubType_DirectDebitOneOff;
        }
        else if (iSubType == eTASubType_DirectDebitRepeat)
          // switching repeat to first
          iSubType = eTASubType_DirectDebitFirst;
      }
      else
      if (szIBAN && strcmp(szIBAN, ptRootMandate->ptNewestVersion->tDebitor.szIBAN) != 0)
      {
        iCreateNew = 1;
        iChange = eMndtChange_IBAN;
      }


      if (strcmp(szDebtName, ptRootMandate->ptNewestVersion->tDebitor.szName) != 0)
      {
        if (iChange != eMndtChange_None)
        {
          fprintf (stderr, "error: transaction #%d: not allowed to change name and account simultaniously\n", ptGlobals->iRecordCount + 1);
          return (0);
        }
        
        iCreateNew = 1;
        iChange = eMndtChange_Name;
        
        fprintf (stderr, "warning: transaction #%d: chaning name of mandate %s to %s\n", ptGlobals->iRecordCount + 1,
                          ptRootMandate->ptNewestVersion->tDebitor.szName, szDebtName);
      }

      // keep this version in mind as latest used version
      if (iCreateNew)
        ptRootMandate->ptLastUsedVersion = ptRootMandate->ptNewestVersion;

    }

    ptGlobals->iRecordCount++;

    if (iSubType != ptGlobals->iWhatToEmit)
    {
      if (    (ptGlobals->iWhatToEmit != eTASubType_DirectDebitEmitDefault)
           || (iSubType != eTASubType_DirectDebitRepeat))
      {

        if (ptGlobals->iWhatToEmit == eTASubType_DirectDebitEmitDefault)
          fprintf (stderr, "warning: transaction #%d: does not match target type\n", ptGlobals->iRecordCount);

        free(szMandate);
        free(szDebtName);
        if (szIBAN)
          free(szIBAN);
        if (szBIC)
          free(szBIC);
        free (szPurpose);


        return (-1);
      }
    }

    if (iCreateNew)
    {
      struct t_Mandate * ptMandate = (struct t_Mandate *) malloc (sizeof(struct t_Mandate));

      ptMandate->szMandate		= szMandate;
      ptMandate->iMndtChanged		= iChange;
      ptMandate->tDebitor.szName	= szDebtName;
      ptMandate->iValidSincePersMsgID	= ptGlobals->iPersistentMsgID;

      if (szIBAN)
        ptMandate->tDebitor.szIBAN	=  szIBAN;
      else
      {
        ptMandate->tDebitor.szIBAN = (char *) malloc (strlen(ptRootMandate->ptLastUsedVersion->tDebitor.szIBAN) + 1);
        strcpy(ptMandate->tDebitor.szIBAN, ptRootMandate->ptLastUsedVersion->tDebitor.szIBAN);
      }

      if (szBIC)
        ptMandate->tDebitor.szBIC	= szBIC;
      else
      {
        ptMandate->tDebitor.szBIC = (char *) malloc (strlen(ptRootMandate->ptLastUsedVersion->tDebitor.szBIC) + 1);
        strcpy(ptMandate->tDebitor.szBIC, ptRootMandate->ptLastUsedVersion->tDebitor.szBIC);
      }

      ptMandate->lDateOfSign		= (lDateOfSign) ? lDateOfSign : ptRootMandate->ptLastUsedVersion->lDateOfSign;


      if ( ! ptRootMandate)
      {
        ptRootMandate = ptMandate;
        ptMandate->ptLastUsedVersion	= NULL;
      }
      ptRootMandate->ptNewestVersion = ptMandate;

      /* chain new entry to the end of the list */
      *ptGlobals->pptMandateListEnd	= ptMandate;
      ptGlobals->pptMandateListEnd	= & ptMandate->ptNext;
      ptMandate->ptNext 		= NULL;
    }
    else
    {
      free(szMandate);
      free(szDebtName);
      if (szIBAN)
        free(szIBAN);
      if (szBIC)
        free(szBIC);
    }
    
    /* new list element */
    struct t_DbtTrfTx * ptNew = (struct t_DbtTrfTx *) malloc (sizeof(struct t_DbtTrfTx));


    /* fill the data */
    ptNew->iType		= iSubType;
    ptNew->ptMandate		= ptRootMandate;
    ptNew->iValueInCent		= iValue;
    ptNew->iSTCount		= 0;
    ptNew->iSTValueInCent	= 0;
    ptNew->szPurpose		= szPurpose;

    ptNew->ptNext		= NULL;
    ptNew->ptNextSubType	= NULL;

    /* adjust the global counters for the elements */
    ptGlobals->iPmtCount++;
    ptGlobals->iPmtSumInCent += iValue;

    /* chain new entry to the end of the list */
    struct t_DbtTrfTx ** ptRun = & ptGlobals->ptTxList.ptDrctDbtTrfTxList;

    while ((*ptRun != NULL) && ((*ptRun)->iType != iSubType))
      ptRun = & (*ptRun)->ptNextSubType;

    /* check type */
    if (*ptRun != NULL)
    {
      /* it's the first subtype entry */
      (*ptRun)->iSTCount++;
      (*ptRun)->iSTValueInCent += iValue;

      /* put the entry at the end of the list */
      while (*ptRun != NULL)
        ptRun = & (*ptRun)->ptNext;
    }
    else
    {
      /* it's the first entry of this subtype */
      ptNew->iSTCount = 1;
      ptNew->iSTValueInCent = iValue;
    }

    /* put the new element at the end of the current list */      
    *ptRun = ptNew;

    /* keep new element as last element in mind */
    ptGlobals->ptLIDrctDbtTrfTx = ptNew;

  }


  return 1;
}


/***********************************************************************************
 * iLineGetCharLength
 *
 * get the number of byte belonging to the character starting with byte iByte
 ***********************************************************************************/
int _iGetCharLength (int iByte)
{
  int iBitMask = 0x80;

  /* check if it is a multibyte char */
  if (iByte & iBitMask)
  {
    /* preset to at least two bytes */
    int iBitCount = 2;

    /* check if the second bit is set (must be in first byte of a multibyte char */
    iBitMask >>= 1;
    if ( ! (iByte & iBitMask))
      /* not set -> return fault */
      return (-1);

    /* starting loop with bit three */
    iBitMask >>= 1;
    /* continue until we have read a zero or if we have examined the entire byte */
    while (iBitMask & iByte)
    {
      /* bit row still set -> continue with next bit */
      iBitCount++;
      iBitMask >>= 1;
    }
    /* first zero found -> return number of bytes to expect */
    return (iBitCount);
  }
  else
    /* 7 bit char -> length equals to one */
    return (1);
}

/***********************************************************************************
 * iGetUnicodeValue
 *
 * converts UTF8 character starting at szChar to a unicode int value
 *
 * parameter
 *   szChar:            pointer to the first char of the UTF-8 code
 *   iMaxLength:        count of valid chars staring and including (*)szChar
 *   piCharLength:      number of UTF-8 chars read (NULL is valid for not required)
 ***********************************************************************************/
int _iGetUnicodeValue (unsigned char * szChar, int iMaxLength, int * piCharLength)
{
  {
    int iReturn = 0;
    int iCurrIndex = 0;
    int iHuntEndOfChar = 0;

    while (((iCurrIndex == 0) || iHuntEndOfChar) && (iCurrIndex < iMaxLength))
    {
      int iByte = szChar[iCurrIndex];
      if (iHuntEndOfChar)
      {
        /* check for following character code */
        if ((iByte & 0xC0) != 0x80)
          return (-1);

        /* get the part of this byte */
        iReturn = (iReturn << 6) | (iByte & 0x3F);

        /* and continue hunting ? */
        iHuntEndOfChar--;
      }
      else
      {
        /* get the length of the current character */
        int iCharLength = _iGetCharLength (iByte);

        /* set return value if desired */
        if (piCharLength)
          *piCharLength = iCharLength;

        /* forward fault */
        if (iCharLength < 0)
          return (-1);

        /* set the first (and probably last) contents of the return value */
        if ((iHuntEndOfChar = (iCharLength - 1)))
          iReturn = iByte & ((1 << (8 - iCharLength)) - 1);
        else
          iReturn = iByte;
      }

      /* go to the next byte */
      iCurrIndex++;
    } /* end-while until end of line */

    /* check if we have reached the position */
    if (iHuntEndOfChar || (iCurrIndex > iMaxLength))
      /* no we are missing some bytes -> fault */
      return (-1);

    return (iReturn);
  }
}

/***********************************************************************************
 ***********************************************************************************/
char           *downcase (char *s)
{
  static char     x[100];
  char           *cp;

  memset (x, 0, sizeof (x));
  strncpy (x, s, 99);

  for (cp = x; *cp; cp++)
    if (isupper (*cp))
      *cp = tolower (*cp);
  return x;
}

/***********************************************************************************
 * returns the first word in this line, returns it and shortens the
 * line.
 ***********************************************************************************/
char           *extract_ident (char *line)
{
  char           *c,
                 *x,
                 *y;
  static char     word[30];

  if (strlen (line) > 0)
  {
    x = strchr (line, ' ');
    y = strchr (line, '\t');

    if (!x && !y)
    {
      strncpy (word, downcase (line), 29);
      line[0] = '\0';
      return word;
    }

    /* Check which index returns the lower value, and check if the
       value is non-NULL */
    if ((c = (x && x < y) ? x : (y ? y : x)))
    {
      strncpy (word, line, c - line);
      word[c - line] = '\0';
      for (; *c == '\t' || *c == ' '; c++);
      for (x = line; *c; c++, x++)
	*x = *c;
      *x = '\0';
      strcpy (word, downcase (word));
      return word;
    }
    return NULL;
  }
  return line;
}

/***********************************************************************************
 ***********************************************************************************/
int rec_index (char *ident, int type)
{
  int             i;
  dtaus_record   *rec = NULL;

  switch (type)
  {
    case REC_A:
      rec = recA;
      break;
    case REC_C:
      rec = recC;
      break;
  }

  for (i = 0; (rec[i].name); i++)
  {
    if (!strcmp (ident, downcase (rec[i].name)))
      return i;
  }

  return -1;
}

/***********************************************************************************
int PASCAL NEAR iGetUnicodeValue (unsigned char * szChar, int iMaxLength, int * piCharLength)
 ***********************************************************************************/
size_t control_nextline (char * szOutBuf, int iOutBufLen, FILE * ptInFile)
{

  int             iOutIndex     = 0;
  int             iLastOutIndex = 0;
  int		  iReadChar;
  int             iCommentLine;
  char *          szReadPointer;


  do
  {
    iReadChar    = 0;
    iCommentLine = 0;
    while (   (iReadChar == 0)
           || (s_szReadBuffer[iReadChar - 1] != '\n'))
    {
      if (iReadChar >= s_iBufferLen - 1)
      {
        s_iBufferLen += 100;
        s_szReadBuffer = realloc(s_szReadBuffer, s_iBufferLen);
      }

      if ( ! fgets(s_szReadBuffer + iReadChar, s_iBufferLen - iReadChar, ptInFile))
        break;

      iReadChar = strlen(s_szReadBuffer);
    }


    szReadPointer = s_szReadBuffer;
    while (iReadChar && (*szReadPointer != '\n'))
    {
      int iCharLength = 0;

      int iChar = _iGetUnicodeValue((unsigned char *) szReadPointer, iReadChar, & iCharLength);

      /* cancel if we have read a non lantin1 char */
      if (iChar > 255)
        return (0);

      szReadPointer += iCharLength;
      iReadChar     -= iCharLength;

      if (iOutIndex < (iOutBufLen - 1))
      {
        /* removing leading whitespaces */
        if ((iOutIndex != 0) || ((iChar != ' ') && (iChar != '\t')))
        {
          if ((iOutIndex == 0) && (iChar == '#'))
            iCommentLine = 1;

          if ( ! iCommentLine)
          {
            szOutBuf[iOutIndex++] = iChar;
            if ((iChar != ' ') && (iChar != '\t'))
              iLastOutIndex = iOutIndex;
          }
        }
      }
    }
  } while ((iOutIndex == 0) && (iReadChar));

  /* remove the trailing whitespaces */
  szOutBuf[iLastOutIndex] = '\0';

  return (szOutBuf[0] != 0);
    
}

/***********************************************************************************
 * Return the current date nicely formatted
 ***********************************************************************************/
char           *get_date ()
{
  static char     res[10];
  time_t          timer;
  struct tm      *loctime;

  timer = time (NULL);
  loctime = localtime (&timer);
  sprintf (res, "%02d.%02d.%02d", loctime->tm_mday, loctime->tm_mon + 1,
	   loctime->tm_year % 100);
  return res;
}

/***********************************************************************************
 ***********************************************************************************/
int control2dtaus (struct t_Globals * ptGlobals,
                   char *ccontrol, char *cdtaus, char *cbeleg, char *ccheck,
		   char *latex)
{
  FILE *	fdtaus		= NULL;
  FILE *	fcontrol;
  FILE *	fbeleg		= NULL;
  FILE *	fcheck		= NULL;

  void           *buf;
  char           *ident;
  int             recindex;
  char            line[100];
  char           *valA[A_LEN],
                 *valC[C_LEN];
  char          **text = NULL;
  char           *cp;
  int             textindex = 0;
  int             len,
                  i;
  char           *type = NULL;
  char           *currency = NULL;
  char            date_todo[11];
  int		  iReturn = 0;

  ptGlobals->ptOutFile = NULL;
  if ( ! (fcontrol = fopen (ccontrol, "r")))
  {
    fclose(fdtaus);
    return 0;
  }

  if ((buf = (char *) malloc (512)) == NULL)
    return 0;

  /* 
   * Record A lesen
   */
  memset (valA, 0, sizeof (valA));
  control_nextline ((void *) line, 100, fcontrol);
  ident = extract_ident (line);
  if (!strcmp (ident, "begin") && (line[0] == '{'))
  {
    control_nextline ((void *) line, 100, fcontrol);
    while (strlen (line) && line[0] != '}')
    {
      ident = extract_ident (line);
      if ((recindex = rec_index (ident, REC_A)) != -1)
      {
	if (recA[recindex].type != IGN)
	  if ((valA[recindex] = (char *) malloc (strlen (line) + 1)) != NULL)
	    strcpy (valA[recindex], line);
      }
      else
      {
	if (!strcasecmp (ident, "BatchBooking"))
	ptGlobals->iBtchBookg = 1;
      }
      control_nextline ((void *) line, 100, fcontrol);
    }

    if (((recindex = rec_index ("art", REC_A)) != -1) && valA[recindex]
	&& strlen (valA[recindex]))
    {
      if (valA[recindex][0] == 'L')
	type = strdup ("Sammeleinziehungsauftrag");
      else if (valA[recindex][0] == 'G')
	type = strdup ("Sammelueberweisungsauftrag");
      else
	type = strdup ("Sammelauftrag");

      currency = strdup ("Euro");

      if (valA[A_WHEN])
	sprintf (date_todo, "%s", valA[A_WHEN]);
      else
	memset (date_todo, 0, sizeof (date_todo));
    }

    if (!dtaus_writeA (ptGlobals, valA))
    {
      fprintf (stderr, "Configuration invalid or not complete\n");
      return 0;
    }

    if ( ! (fdtaus = fopen (cdtaus, "w")))
    {
      fclose(fcontrol);
      return 0;
    }
    ptGlobals->ptOutFile = fdtaus;

    if (cbeleg[0])
    {
      if ( ! (fbeleg = fopen (cbeleg, "w")))
      {
        fclose(fcontrol);
        fclose(fdtaus);
        return 0;
      }
    }

    if (fbeleg)
    {
      fprintf (fbeleg, "\n\n");
      fprintf (fbeleg, "\n    Begleitzettel\n\n");
      fprintf (fbeleg, "\n    Belegloser Datentraegeraustausch\n\n");
      fprintf (fbeleg, "\n    %s\n\n", type);
      fprintf (fbeleg, "\n    VOL ........................: MID-%08x\n",	ptGlobals->iPersistentMsgID);
      fprintf (fbeleg, "\n    Erstellungsdatum ...........: %s\n",		get_date ());
      if (strlen (date_todo))
      {
        fprintf (fbeleg, "\n    Ausfuehrungsdatum ..........: %s\n",
                 date_todo);
      }
      fprintf (fbeleg, "\n    Waehrung ...................: %s\n", currency);
    }
    
    if (ccheck[0])
    {
      if ( ! (fcheck = fopen (ccheck, "w")))
      {
        fclose(fdtaus);
        fclose(fcontrol);
        if (fbeleg)
          fclose(fbeleg);
        return 0;
      }
    }

    if (fcheck)
    {
      fprintf (fcheck, "\n\n\n");
      if ((ptGlobals->iAction == eAction_DirectDebitCore) || (ptGlobals->iAction == eAction_DirectDebitB2B))
        fprintf (fcheck,	"    Sammeleinziehungsauftrag\n\n");
      else if (ptGlobals->iAction == eAction_CreditTransfer)
        fprintf (fcheck,	"    Sammelueberweisungsauftrag\n\n");
      else
        fprintf (fcheck,	"    ungültiger Aufgrag\n\n");
      fprintf (fcheck,		"    Erstellungsdatum : %s\n\n", get_date ());
      fprintf (fcheck,		"    Waehrung         : Euro\n\n\n");
      fprintf (fcheck,		"     %-22s  %-11s  %-30s   %12s\n", "IBAN", "BIC", "Name", "Betrag");
      fprintf (fcheck,		"    ------------------------------------------------------------------------------------\n");
    }
  }
  else
  {
    fprintf (stderr,
	     "configuration block (BEGIN) not found in configuration file\n");
    return 0;
  }

  /* 
   * Record C lesen
   */
  memset (valC, 0, sizeof (valC));
  control_nextline ((void *) line, 100, fcontrol);
  if (line[0] == '{')
  {
    while (strlen (line) && line[0] == '{')
    {
      control_nextline ((void *) line, 100, fcontrol);
      if (text)
      {
	for (textindex = 0; textindex < MAX_TEXT && text[textindex]; textindex++)
	{
	  if (text[textindex])
	  {
            free (text[textindex]);
            text[textindex] = NULL;
          }
        }
      }
      else
      {
        if ((text = (char **) malloc ((MAX_TEXT + 1) * sizeof (char *))) == NULL)
          return 0;
        else
          memset (text, 0, (MAX_TEXT + 1) * sizeof (char *));
      }
      textindex = 1;

      while (strlen (line) && line[0] != '}')
      {
	ident = extract_ident (line);
	if ((recindex = rec_index (ident, REC_C)) != -1)
	  if (recC[recindex].type != IGN)
	  {
	    /*
	     * Special exception to support multiple Text fields
	     */
	    if (recindex == C_TEXT && valC[recindex])
	    {
	      if (textindex < MAX_TEXT)
	      {
		if ((cp = (char *) malloc (strlen (line) + 1)) == NULL)
		  return 0;
		strcpy (cp, line);
		cp[strlen (line)] = '\0';
		text[textindex++] = cp;
	      }
	    }
	    else
	    {
	      len = strlen (line);
	      if (recindex == C_VAL)
	      {
		/* Convert commas to dots for later processing */
		for (i = 0; line[i]; i++)
		  if (line[i] == ',')
		    line[i] = '.';

		if ((cp = strchr (line, '.')) == NULL)
		{
		  if (strlen (line) > 9)
		  {
		    fprintf (stderr, "Betrag %s zu gross (max. 9 Stellen)\n",
			     line);
		    return 0;
		  }
		  if ((valC[recindex] =
		       (char *) malloc (strlen (line) + 4)) == NULL)
		    return 0;
		  sprintf (valC[recindex], "%s.00", line);
		}
		else if (((len = cp - line + 3)) < strlen (line))
		{
		  if (cp - line > 9)
		  {
		    fprintf (stderr,
			     "Betrag %s zu gross (max. 9.2 Stellen)\n", line);
		    return 0;
		  }
		  if ((valC[recindex] = (char *) malloc (len + 1)) == NULL)
		    return 0;
		  strncpy (valC[recindex], line, len);
		  valC[recindex][len] = '\0';
		}
		else
		{
		  if (cp - line > 9)
		  {
		    fprintf (stderr,
			     "Betrag %s zu gross (max. 9.2 Stellen)\n", line);
		    return 0;
		  }
		  if ((valC[recindex] =
		       (char *) malloc (strlen (line) + 1)) == NULL)
		    return 0;
		  strcpy (valC[recindex], line);
		}
	      }
	      else
	      {
		if ((valC[recindex] =
		     (char *) malloc (strlen (line) + 1)) == NULL)
		  return 0;
		strcpy (valC[recindex], line);
	      }
	    }
	  }
	control_nextline ((void *) line, 100, fcontrol);
      }

      {
        int iRetVal;
        if ((iRetVal = dtaus_writeC (ptGlobals, valA, valC, text)) == 0)
          return 0;
        else if ((iRetVal > 0) && fcheck)
        {
          int			iValueInCent;
          struct t_CredDebt *	ptPayment;

	  if (_GetLastInserted(ptGlobals, & iValueInCent, & ptPayment))
            fprintf (fcheck, "     %10s  %8s  %-30s   %12s\n",
				   ptPayment->szIBAN, ptPayment->szBIC,
				   ptPayment->szName,
				   GetValueFromCentString(iValueInCent));
        }
      }

      for (recindex = 0; recindex < C_LEN; recindex++)
	if (valC[recindex])
	  free (valC[recindex]);
      memset (valC, 0, sizeof (valC));
      control_nextline ((void *) line, 100, fcontrol);
    }
    switch(ptGlobals->iAction)
    {
      case eAction_CreditTransfer:
        iReturn = iDoCreditTransfer(ptGlobals);
        break;

      case eAction_DirectDebitCore:
        iReturn = iDoDirectDebitInitiation(ptGlobals);
        break;

      default:
        fprintf(stderr, "action is currently not supported\n");
    }

  }
  else
  {
    fprintf (stderr, "Kein regulärer Datensatz?\n");
    return 0;
  }

  if (text)
  {
    for (textindex = 0; textindex < MAX_TEXT && text[textindex]; textindex++)
    {
      if (text[textindex])
        free (text[textindex]);
    }
    free(text);
  }


  if (fbeleg)
  {
    fprintf (fbeleg, "\n    Anzahl .....................: %d\n", ptGlobals->iPmtCount);
    fprintf (fbeleg, "\n    Summe ......................: %s\n", GetValueFromCentString(ptGlobals->iPmtSumInCent));
    fprintf (fbeleg, "\n    Unsere IBAN ................: %s\n", ptGlobals->ptInitiator->szIBAN);
    fprintf (fbeleg, "\n    Unsere BIC .................: %s\n", ptGlobals->ptInitiator->szBIC);
    fprintf (fbeleg,
             "\n\n\n\n\n    __________________________________________________\n");
    fprintf (fbeleg, "    Ort, Datum                     Unterschrift\n");
  }

  if (fcheck)
  {
    fprintf (fcheck,
             "    ------------------------------------------------------------------------------------\n");
    fprintf (fcheck, "     %-67s %14s\n", "Summe", GetValueFromCentString(ptGlobals->iPmtSumInCent));
  }

  _GiveUpTransactions(ptGlobals);
  _GiveUpInitiator(ptGlobals);

  free (buf);

  if (s_iBufferLen)
  {
    free(s_szReadBuffer);
    s_iBufferLen = 0;
    s_szReadBuffer = NULL;
  }
  
  for (recindex = 0; recindex < A_LEN; recindex++)
    if (valA[recindex])
      free (valA[recindex]);
  fclose (fdtaus);
  fclose (fcontrol);
  if (fbeleg)
    fclose (fbeleg);
  if (fcheck)
    fclose (fcheck);

  return (iReturn);
}



/***********************************************************************************
 ***********************************************************************************/
int _GetTransactionSubType(struct t_Globals * ptGlobals, char * szType)
{
  struct { char * szName; enum eTASubType iType; } 		atTypeMapping[] = {
    { "Gutschrift",		eTASubType_Credit},
    { "Lohn",			eTASubType_Salary},
    { "Einzug-Erster",		eTASubType_DirectDebitFirst},
    { "Einzug-Wiederholung",	eTASubType_DirectDebitRepeat},
    { "Einzug-Letzter",		eTASubType_DirectDebitFinal},
    { "Einzug-Einmalig",	eTASubType_DirectDebitOneOff},
    { NULL,			0}},				* ptTypeMapping;

  ptTypeMapping = atTypeMapping;


  while (ptTypeMapping->szName)
  {
    if (strcmp(ptTypeMapping->szName, szType) == 0)
    {
      if (    (    (ptGlobals->iAction == eAction_CreditTransfer)
                && (ptTypeMapping->iType & eTASubType_CreditTransferType))
           || (    (    (ptGlobals->iAction == eAction_DirectDebitCore)
                     || (ptGlobals->iAction == eAction_DirectDebitB2B))
                && (ptTypeMapping->iType & eTASubType_DebitTransferType)))
        return (ptTypeMapping->iType);
    }
    ptTypeMapping++;
  }


  return (0);
}

/***********************************************************************************
 ***********************************************************************************/
static struct  t_CredDebt * _GetCredDebt(char ** pszName, char ** pszIBAN, char ** pszBIC)
{
  if ( ! _CheckIBAN(*pszIBAN))
    return (NULL);

  if ( ! _CheckBIC(*pszBIC))
    return (NULL);

  struct  t_CredDebt * ptReturn = (struct t_CredDebt *) malloc(sizeof(struct t_CredDebt));
  ptReturn->szName		= _GetISOString(*pszName, NULL, 70, 0);
  ptReturn->szIBAN		= *pszIBAN;
  ptReturn->szBIC		= *pszBIC;

  *pszIBAN = NULL;
  *pszBIC  = NULL;

  return (ptReturn);
}

/***********************************************************************************
 ***********************************************************************************/
static int _CheckBIC(char * szBIC)
{
  /* valid BIC [A-Z]{6,6}[A-Z2-9][A-NP-Z0-9] ([A-Z0-9]{3,3}){0,1} */
  unsigned char szBICString[] = { 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
                                  0x04, 0x08,
                                  0x03, 0x03, 0x03};
  int iI;

  for (iI = 0; iI < strlen(szBIC) && iI < 11; iI++)
  {
    if (szBIC[iI] > 127)
      break;

    if ((szCharClass[(unsigned int) szBIC[iI]] & szBICString[iI]) == 0)
      break;
  }

  if ((szBIC[iI] == '\0') && ((iI == 8) || (iI == 11)))
    return (1);

  return (0);
}


/***********************************************************************************
 ***********************************************************************************/
int _CheckIBAN(char * szIBAN)
{
  
  /* valid IBAN [A-Z]{2,2}[0-9]{2,2}[a-zA-Z0-9]{1,30} */
  unsigned char szIBANString[] = { 0x01, 0x01,
                                   0x02, 0x02,
                                   0x13, 0x13, 0x13, 0x13, 0x13, 0x13, 0x13, 0x13, 0x13, 0x13,
                                   0x13, 0x13, 0x13, 0x13, 0x13, 0x13, 0x13, 0x13, 0x13, 0x13,
                                   0x13, 0x13, 0x13, 0x13, 0x13, 0x13, 0x13, 0x13, 0x13, 0x13};
  int		iI;
  int		iIBANLength	= strlen(szIBAN);
  char		szCheckBuffer[100];
  int		iCheckIndex	= 0;

  for (iI = 4;  (iI < iIBANLength + 4) && (iI < 38); iI++)
  {
    int		iStringIndex	= (iI >= iIBANLength) ? iI - iIBANLength : iI;
    int		iChar		= (unsigned char) szIBAN[iStringIndex];
    
    if (iChar > 127)
      break;

    if ((szCharClass[iChar] & szIBANString[iStringIndex]) == 0)
      break;

    if (iChar < 'A')
      szCheckBuffer[iCheckIndex++] = (unsigned char) iChar;
    else
    {
      iChar = toupper(iChar) - 'A' + 10;
      szCheckBuffer[iCheckIndex++] = iChar / 10 + '0';
      szCheckBuffer[iCheckIndex++] = iChar % 10 + '0';
    }
  }

  if ((iI != (iIBANLength + 4)) || (iIBANLength <= 4))
    return (0);

  // terminate the checksum buffer iCheckIndex stay constant now
  szCheckBuffer[iCheckIndex] = '\0';

  char *	szCheckStart	= szCheckBuffer;
  char *	szEndPtr	= szCheckBuffer;
  char		cReplacedChar;
  long		lCheckValue;

  while (1)
  {

    cReplacedChar = szCheckStart[9];
    szCheckStart[9] = '\0';
    lCheckValue = strtol(szCheckStart, & szEndPtr, 10);

    lCheckValue = lCheckValue % 97;

    if (szEndPtr == szCheckBuffer + iCheckIndex)
      break;

    szCheckStart = szEndPtr;
    *(szCheckStart) = cReplacedChar;
    *(--szCheckStart) = lCheckValue % 10 + '0';
    if (lCheckValue / 10)
      *(--szCheckStart) = lCheckValue / 10 + '0';
  }
  
  return (lCheckValue == 1);
}


/***********************************************************************************
 ***********************************************************************************/
int _GetValueInCent(char * szValue)
{

  char *	szRunning	= szValue;
  char *	szValuePart;
  char *	szValueEndPtr;
  long 		lValueEuros;
  long		lValueCents;

  szValuePart = strsep( & szRunning, ".,");
  lValueEuros = strtol(szValuePart, & szValueEndPtr, 10);
  /* we have to have another part to parse and first part has to be entirely converted */
  if (szRunning && (*szValuePart != '\0') && (*szValueEndPtr == '\0'))
  {
    szValuePart = strsep( & szRunning, ".,");
    lValueCents = strtol(szValuePart, & szValueEndPtr, 10);
    /* we do not have to have another part to parse and second part has to be entirely converted */
    if ( ! szRunning && (*szValuePart != '\0') && (*szValueEndPtr == '\0'))
      /* valid szValue string */
      return ((int) (lValueEuros * 100 + lValueCents));
  }

  /* return invalid amount */
  return (-1);
}


/***********************************************************************************
 ***********************************************************************************/
char * _GetPurposeString(char * szPurpose, int iPurposeLen, int iAddTextLines, char ** aszText)
{
  char *	szReturn	= NULL;
  char *	szRun;
  char *	szLastLetter;
  int		iNextText	= 0;

  // get the first part
  szReturn = _GetISOString(szPurpose, szReturn, iPurposeLen, 0);
  if ( ! szReturn)
    return (NULL);

  szRun = szReturn + strlen(szReturn);
  szLastLetter = szReturn + iPurposeLen;

  while ((szRun < (szLastLetter - 1)) && aszText && (iNextText < iAddTextLines))
  {
    *(szRun++) = ' ';
    if ( ! _GetISOString(aszText[iNextText++], szRun, szLastLetter - szRun, 0))
    {
      free (szReturn);
      return (NULL);
    }

    szRun += strlen(szRun);
  }

  if (aszText && (iNextText < iAddTextLines))
  {
      free (szReturn);
      return (NULL);
  }

  return (szReturn);
}

/***********************************************************************************
 ***********************************************************************************/
char * _GetISOString(char * szSource, char * szDest, int iMaxDestLen, int iUpperCase)
{
  int		iRemainingLen	= iMaxDestLen;
  char *	szReturn;
  char * 	szOutPtr;

  szReturn = (szDest) ? szDest : (char *) malloc (iMaxDestLen + 1);
  szOutPtr = szReturn;


  while ( *szSource && (iRemainingLen > 0))
  {
    int iCurrChar = *((unsigned char *) szSource++);
    if ((iCurrChar < 128) && (szCharClass[iCurrChar] & 0x80))
    {
      if (iUpperCase)
        *(szOutPtr++) = toupper(iCurrChar);
      else
        *(szOutPtr++) = iCurrChar;

      iRemainingLen--;
    }
    else
    {
      char * szReplacement = NULL;

      switch (iCurrChar)
      {
        case 0xE4:
          szReplacement = "ae";
          break;
        case 0xC4:
          szReplacement = "Ae";
          break;
        case 0xF6:
          szReplacement = "oe";
          break;
        case 0xD6:
          szReplacement = "Oe";
          break;
        case 0xFC:
          szReplacement = "ue";
          break;
        case 0xDC:
          szReplacement = "Ue";
          break;
        case 0xDF:
          szReplacement = "ss";
          break;
      }


      if (( ! szReplacement) || (strlen(szReplacement) > iRemainingLen))
      {
        if ( ! szDest)
          free (szReturn);

        return (NULL);
      }
      else
      {
        while (*szReplacement)
        {
          if (iUpperCase)
            *(szOutPtr++) = toupper(*((unsigned char *) szReplacement++));
          else
            *(szOutPtr++) = *(szReplacement++);
        }
      }
    } /* end-else special char */
  } /* end-while through the source string */

  // check if entire string was converted
  if (*szSource)
  {
    if ( ! szDest)
      free (szReturn);

    return (NULL);
  }

  *szOutPtr = '\0';

  return (szReturn);
}


/***********************************************************************************
 ***********************************************************************************/
time_t _GetISODate(char * szDate, int iCheckPast)
{
  char * 	szRunning = szDate;
  int    	iError = 1;
  time_t	iReturn;
  struct tm	tTime;
  struct tm	tCurrTime;
  time_t	iCurrTime = time(NULL);

  localtime_r( & iCurrTime, & tCurrTime);
  tCurrTime.tm_isdst = 0;
  tCurrTime.tm_gmtoff = 0;
  tCurrTime.tm_sec = 0;
  tCurrTime.tm_min = 0;
  tCurrTime.tm_hour = 0;
  iCurrTime = mktime( & tCurrTime);
  
  tTime.tm_isdst = 0;
  tTime.tm_gmtoff = 0;
  tTime.tm_sec  = 0;
  tTime.tm_min  = 0;
  tTime.tm_hour = 0;

  tTime.tm_mday = atoi(strsep( & szRunning, "."));
  if (szRunning)
  {
    *(szRunning - 1) = '.';
    tTime.tm_mon  = atoi(strsep( & szRunning, ".")) - 1;
    if (szRunning)
    {
      *(szRunning - 1) = '.';
      tTime.tm_year = atoi(strsep( & szRunning, ".")) - 1900;
      iError = ! ((tTime.tm_mon >= 0) && (tTime.tm_year >= 0));
    }
  }

  if (iError || ((iReturn = mktime( & tTime)) == -1) || (iCheckPast && (iReturn < iCurrTime)))
  {
    return (0);
  }

  return (iReturn + tTime.tm_gmtoff);
}


/***********************************************************************************
 ***********************************************************************************/
int _GiveUpTransactions(struct t_Globals * ptGlobals)
{
  switch(ptGlobals->iAction)
  {
    case eAction_CreditTransfer:
    {
      struct t_CdtTrfTx * ptTxList = ptGlobals->ptTxList.ptCdtTrfTxList;

      while (ptTxList)
      {
        
        struct t_CdtTrfTx *	ptToDelete	= ptTxList;
        ptTxList				= ptTxList->ptNext;

        free(ptToDelete->ptCreditor->szName);
        free(ptToDelete->ptCreditor->szIBAN);
        free(ptToDelete->ptCreditor->szBIC);

        free (ptToDelete->ptCreditor);

        free(ptToDelete->szPurpose);

        free(ptToDelete);
        
      }
      break;
    }

    case eAction_DirectDebitCore:
    {
      struct t_DbtTrfTx * ptTxList = ptGlobals->ptTxList.ptDrctDbtTrfTxList;

      while (ptTxList)
      {
        struct t_DbtTrfTx *	ptTxSubList	= ptTxList;
        ptTxList				= ptTxList->ptNextSubType;

        while (ptTxSubList)
        {
          struct t_DbtTrfTx * ptToDelete = ptTxSubList;
          ptTxSubList = ptTxSubList->ptNext;

          free(ptToDelete->szPurpose);
          free(ptToDelete);
        }
      }
      break;
    }

    default:
      return (0);
  }
  return (1);
}


/***********************************************************************************
 ***********************************************************************************/
int _GiveUpInitiator(struct t_Globals * ptGlobals)
{
  if (ptGlobals->ptInitiator)
  {
    if (ptGlobals->ptInitiator->szName)
      free (ptGlobals->ptInitiator->szName);
    if (ptGlobals->ptInitiator->szIBAN)
      free (ptGlobals->ptInitiator->szIBAN);
    if (ptGlobals->ptInitiator->szBIC)
      free (ptGlobals->ptInitiator->szBIC);

    free (ptGlobals->ptInitiator);
    return (1);
  }
  return (0);
}


/***********************************************************************************
 ***********************************************************************************/
int _GetLastInserted(	struct t_Globals * ptGlobals,
			int * piValueInCent, struct t_CredDebt ** pptPayment)
{
  if (ptGlobals->ptLICdtTrfTx)
  {
    *piValueInCent	= ptGlobals->ptLICdtTrfTx->iValueInCent;
    *pptPayment		= ptGlobals->ptLICdtTrfTx->ptCreditor;
  }
  else if (ptGlobals->ptLIDrctDbtTrfTx)
  {
    *piValueInCent	= ptGlobals->ptLIDrctDbtTrfTx->iValueInCent;
    *pptPayment 	= & ptGlobals->ptLIDrctDbtTrfTx->ptMandate->ptNewestVersion->tDebitor;
  }
  else
    return (0);

  return (1);
}

