/*
    main.c - Belegloser Datenträgeraustausch mit einer Bank
    Copyright (c) 1996,8,2001-4  Martin Schulze <joey@infodrom.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111, USA.

    $Id: main.c,v 1.20 2004-08-06 21:14:06 joey Exp $
 */

static char szVersion[] = "0.4.0";

#include "config.h"

#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>

#include "sepa.h"
#include "dtaus.h"

#include "memory.h"

/****************************************************************************************
 ****************************************************************************************/
#define FN_DTAUS   	"sepa0.ctl"
#define FN_DATABASE	".sepa"
#define FS_XML		".xml"
#define FS_RECEIPT	".doc"
#define FS_LOG		".sik"


enum eGlobalAction
{
  eGlobalAction_CreateXML,
  eGlobalAction_ShowAllMndts,
  eGlobalAction_ShowMandate,
  eGlobalAction_ShowVersion,
  eGlobalAction_Undo,
};

/****************************************************************************************
 ****************************************************************************************/
static void 	_help (char * szControlFileName, char * szDatabaseFileName, char * szReceiptFileName, char * szLogFileName);
static int	_WriteDatabase(struct t_Globals * ptGlobals, char * szFileName);
static int	_ReadDatabase(struct t_Globals * ptGlobals, char * szFileName);
static int	_GiveUpDatabase(struct t_Globals * ptGlobals);

static int	_ShowAllMndts(struct t_Globals * ptGlobals);
static int	_ShowMandate(struct t_Globals * ptGlobals, char * szMandateID);


/****************************************************************************************
 ****************************************************************************************/
enum eDBType
{
  eDBType_BaseTypeInt		= 0x01000000,
  eDBType_BaseTypeLong		= 0x02000000,
  eDBType_BaseTypeString	= 0x04000000,
  eDBType_BaseType		= 0x07000000,
  eDBType_Type			= 0x7FFF0000,
  eDBType_EntryMaskLen		= 16,


  eDBType_TypePersistentMsgID	= 0x00000000 | eDBType_BaseTypeInt, 
  eDBType_TypeMandateName	= 0x00010000 | eDBType_BaseTypeString, 
  eDBType_TypeDateOfSign	= 0x00020000 | eDBType_BaseTypeLong, 
  eDBType_TypeName		= 0x00040000 | eDBType_BaseTypeString, 
  eDBType_TypeIBAN		= 0x00050000 | eDBType_BaseTypeString, 
  eDBType_TypeBIC		= 0x00060000 | eDBType_BaseTypeString, 
  eDBType_TypeValidSince	= 0x00070000 | eDBType_BaseTypeInt, 

};


/****************************************************************************************
 ****************************************************************************************/
void _help (char * szControlFileName, char * szDatabaseFileName, char * szReceiptFileName, char * szLogFileName)
{
  fprintf(stderr, "usage: sepa [-c <control file>] [-d <database file>] [-r] [-l]\n");
  fprintf(stderr, "       sepa [-s] [-d <database file>] \n");
  fprintf(stderr, "       sepa [-m <mandate id>] [-d <database file>] \n");
  fprintf(stderr, "       sepa [-v]\n");
  fprintf (stderr, "\n");
  fprintf (stderr, "  -c <control file>   control data         [%s]\n",	szControlFileName);
  fprintf (stderr, "  -d <database file>  mandate database     [%s]\n",	szDatabaseFileName);
  fprintf (stderr, "  -t [flro]           debit type to filter [r]\n");
  fprintf (stderr, "  -r                  create receipt       [%s]\n",	szReceiptFileName);
  fprintf (stderr, "  -l                  create log file      [%s]\n",	szLogFileName);
  fprintf (stderr, "  -s                  show all mandates\n");
  fprintf (stderr, "  -m <mandate id>     show particular mandate\n");
  fprintf (stderr, "  -v                  show version\n");
}

/****************************************************************************************
 ****************************************************************************************/
int main (int argc, char **argv)
{
  char			szControlFileName[FILENAME_MAX];
  char			szDatabaseFileName[FILENAME_MAX];
  char			szReceiptFileName[FILENAME_MAX];
  char			szLogFileName[FILENAME_MAX];
  char			szXMLFileName[FILENAME_MAX];

  char *		szTmp;

  struct t_Globals	tGlobals;
  struct t_CredDebt	tInitiator;
  int			iAction		= eGlobalAction_CreateXML;


  tGlobals.ptInitiator			= & tInitiator;
  tGlobals.ptOutFile			= NULL;
  tGlobals.iPersistentMsgID		= 0;
  tGlobals.iBtchBookg			= 0;

  tGlobals.iExecTimeStamp		= 0;
  
  tGlobals.szIdentification		= NULL;
  tGlobals.ptInitiator			= NULL;
  tGlobals.ptCreditor			= NULL;
  tGlobals.ptDebitor			= NULL;

  tGlobals.ptMandateList		= NULL;
  tGlobals.pptMandateListEnd		= &  tGlobals.ptMandateList;

  tGlobals.ptTxList.ptCdtTrfTxList	= NULL;

  tGlobals.ptLICdtTrfTx			= NULL;
  tGlobals.ptLIDrctDbtTrfTx		= NULL;

  tGlobals.iRecordCount			= 0;
  tGlobals.iPmtCount			= 0;
  tGlobals.iPmtSumInCent		= 0;

  tGlobals.iWhatToEmit			= eTASubType_DirectDebitEmitDefault;
  

  strcpy(szControlFileName, FN_DTAUS);
  strcpy(szReceiptFileName, szControlFileName);
  strcpy(szLogFileName, szControlFileName);

  char * ptSep;
  int	 iSepPos;
  if ((ptSep = strrchr(szControlFileName, '.')))
    iSepPos = ptSep - szControlFileName;
  else
    iSepPos = strlen(szControlFileName);

  strcpy(szReceiptFileName + iSepPos, FS_RECEIPT);
  strcpy(szLogFileName + iSepPos, FS_LOG);
  
  if ( ! (szTmp = getenv("HOME")))
    exit (1);
  strncpy(szDatabaseFileName, szTmp, FILENAME_MAX);
  szTmp = szDatabaseFileName + strlen(szTmp);
  strcpy(szTmp, "/" FN_DATABASE);

  int		iOpt;
  int		iDoReceipt = 0;
  char *	szMandateID = NULL;
  int		iDoLog = 0;

  while ((iOpt = getopt(argc, argv, "c:d:rlvsm:ut:")) != -1)
  {
    switch (iOpt)
    {
      case 'c':
      {
        strcpy(szControlFileName, optarg);
        break;
      }

      case 'd':
      {
        strcpy(szDatabaseFileName, optarg);
        break;
      }

      case 'r':
      {
        iDoReceipt = 1;
        break;
      }

      case 'l':
      {
        iDoLog = 1;
        break;
      }

      case 's':
      {
        iAction = eGlobalAction_ShowAllMndts;
        break;
      }

      case 'm':
      {
        iAction = eGlobalAction_ShowMandate;
        szMandateID = (char *) malloc (strlen(optarg) + 1);
        strcpy(szMandateID, optarg);

        break;
      }

      case 'u':
      {
        iAction = eGlobalAction_Undo;
        break;
      }

      case 'v':
      {
        iAction = eGlobalAction_ShowVersion;
        break;
      }

      case 't':
      {
        char *			szIndex;
        char *			szDelim		= "flro";
        enum eTASubType		aiTypes[] 	= {	eTASubType_DirectDebitFirst,
							eTASubType_DirectDebitFinal,
							eTASubType_DirectDebitRepeat,
							eTASubType_DirectDebitOneOff
						  };
        if ((strlen(optarg) != 1) || ((szIndex = index("flro", *optarg)) == NULL))
        {
          _help(szControlFileName, szDatabaseFileName, szReceiptFileName, szLogFileName);
          exit(1);
        }
        
        tGlobals.iWhatToEmit = aiTypes[szIndex - szDelim];
        break;
      }


      default: /* '?' */
        _help(szControlFileName, szDatabaseFileName, szReceiptFileName, szLogFileName);
        exit(1);  
    }
  }

  strcpy(szXMLFileName, szControlFileName);
  strcpy(szReceiptFileName, szControlFileName);
  strcpy(szLogFileName, szControlFileName);

  if ((ptSep = strrchr(szControlFileName, '.')))
    iSepPos = ptSep - szControlFileName;
  else
    iSepPos = strlen(szControlFileName);

  strcpy(szXMLFileName + iSepPos, FS_XML);
  if (iDoReceipt)
    strcpy(szReceiptFileName + iSepPos, FS_RECEIPT);
  else
    szReceiptFileName[0] = '\0';

  if (iDoLog)
    strcpy(szLogFileName + iSepPos, FS_LOG);
  else
    szLogFileName[0] = '\0';


  if ( ! _ReadDatabase( & tGlobals, szDatabaseFileName))
  {
    fprintf(stderr, "error: couldnt read database\n");
    exit (1);
  }


  switch(iAction)
  {
    case eGlobalAction_CreateXML:
    {
      struct stat		sbuf;
      if ( ! stat(szControlFileName, &sbuf))
      {
        if (control2dtaus ( & tGlobals, szControlFileName, szXMLFileName, szReceiptFileName, szLogFileName, NULL) == 0)
          exit (1);
      }
      else
      {
        fprintf( stderr, "Can't stat %s\n", szControlFileName);
        exit (1);
      }
      tGlobals.iPersistentMsgID++;
      _WriteDatabase( & tGlobals, szDatabaseFileName);

      break;
    }

    case eGlobalAction_ShowAllMndts:
    {
      _ShowAllMndts( & tGlobals);
      break;
    }

    case eGlobalAction_ShowMandate:
    {
      _ShowMandate( & tGlobals, szMandateID);
      break;
    }

    case eGlobalAction_ShowVersion:
    {
      printf("sepa version: %s\n", szVersion);
      break;
    }

    case eGlobalAction_Undo:
    {
      tGlobals.iPersistentMsgID--;
      _WriteDatabase( & tGlobals, szDatabaseFileName);
      break;
    }



  }

  if (szMandateID)
    free (szMandateID);

  _GiveUpDatabase( & tGlobals);

  return (0);
}


/****************************************************************************************
 *
 *
 * return
 *   bool		<success>
 ****************************************************************************************/
int _ShowAllMndts(struct t_Globals * ptGlobals)
{
  struct t_Mandate *	ptCurrMandate	= ptGlobals->ptMandateList;

  while (ptCurrMandate)
  {
    struct t_Mandate *	ptMandateToShow	= (ptCurrMandate->ptNewestVersion)
						? ptCurrMandate->ptNewestVersion
						: ptCurrMandate;

    if (ptMandateToShow->iValidSincePersMsgID)
    {
      printf("MandateID: %s\n", ptMandateToShow->szMandate);
      printf("         Name: %s\n", ptMandateToShow->tDebitor.szName);
      printf("         IBAN: %s\n", ptMandateToShow->tDebitor.szIBAN);
      printf("          BIC: %s\n", ptMandateToShow->tDebitor.szBIC);
      printf("  last change: %d runs ago\n\n", ptGlobals->iPersistentMsgID - ptMandateToShow->iValidSincePersMsgID);

      ptMandateToShow->iValidSincePersMsgID = 0;
    }

    ptCurrMandate = ptCurrMandate->ptNext;
  }

  return (1);
}



/****************************************************************************************
 *
 *
 * return
 *   bool		<success>
 ****************************************************************************************/
static int	_ShowMandate(struct t_Globals * ptGlobals, char * szMandateID)
{
  struct t_Mandate *	ptCurrMandate	= ptGlobals->ptMandateList;
  int			iFound		= 0;

  while (ptCurrMandate)
  {
    if (strcmp(ptCurrMandate->szMandate, szMandateID) == 0)
    {
      if (iFound == 0)
      {
        printf("Mandate: %s\n", szMandateID);
        printf ("  %-30s   %-22s  %-11s    %s\n", "IBAN", "BIC", "Name", "last change");
        iFound = 1;
      }

      printf ("   %-30s   %-22s  %-11s %9d ago\n", ptCurrMandate->tDebitor.szName, ptCurrMandate->tDebitor.szIBAN, ptCurrMandate->tDebitor.szBIC, ptGlobals->iPersistentMsgID - ptCurrMandate->iValidSincePersMsgID);

    }

    ptCurrMandate = ptCurrMandate->ptNext;
  }

  return (iFound != 0);
}


/****************************************************************************************
 *
 *
 * return
 *   bool		<success>
 ****************************************************************************************/
int _ReadDatabase(struct t_Globals * ptGlobals, char * szFileName)
{
  FILE * ptDataBaseFile = fopen(szFileName, "r");
  if (ptDataBaseFile)
  {
    ptGlobals->pptMandateListEnd		= & ptGlobals->ptMandateList;
    struct t_Mandate *	ptCurrMandate		= NULL;
    struct t_Mandate *	ptLastNewMandate	= NULL;
    struct t_Mandate *	ptLastNewestVersion	= NULL;

    fseek(ptDataBaseFile, 0, SEEK_END);
    long lFileLen = ftell(ptDataBaseFile);
    fseek(ptDataBaseFile, 0, SEEK_SET);

    long lBytesToRead = lFileLen;

    while (lBytesToRead)
    {
      union { int iInt; unsigned long lLong; char * szString; }
						tValue;
      int					iType;
      void * 					vptValue	= NULL;
      long 					lLenItem	= sizeof(iType);
      long					lFreadLen;

      if (lBytesToRead < lLenItem)
        break;

      if ((lFreadLen = fread( & iType, lLenItem, 1, ptDataBaseFile)) != 1)
      {
        fprintf(stderr, "%d: >%s<\n", (int) lFreadLen, strerror(errno));
        break;
      }
      lBytesToRead -= lLenItem;

      lLenItem = (iType & ((1 << eDBType_EntryMaskLen) - 1));

      switch(iType & eDBType_BaseType)
      {
        case eDBType_BaseTypeInt:
          vptValue =  & tValue.iInt;
          break;

        case eDBType_BaseTypeLong:
          vptValue =  & tValue.lLong;
          break;

        case eDBType_BaseTypeString:
          tValue.szString = malloc(lLenItem);
          vptValue = tValue.szString;
          break;
      }

      if ( ! vptValue)
        break;

      if (fread(vptValue, lLenItem, 1, ptDataBaseFile) != 1)
        break;
      lBytesToRead -= lLenItem;
      
      switch(iType & eDBType_Type)
      {
        case eDBType_TypePersistentMsgID:
          ptGlobals->iPersistentMsgID = tValue.iInt;
          break;

        case eDBType_TypeMandateName:
          if (ptCurrMandate)
            iType = 0;
          else
          {
            ptCurrMandate = (struct t_Mandate *) malloc (sizeof(struct t_Mandate));
            ptCurrMandate->szMandate	= tValue.szString;

            ptCurrMandate->iMndtChanged	= 0;
            ptCurrMandate->ptNext		= NULL;
            ptCurrMandate->ptNewestVersion	= ptCurrMandate;
            ptCurrMandate->ptLastUsedVersion	= NULL;

            ptLastNewestVersion = NULL;

            /* look if we have an older version of this mandate in the list */
            {
              struct t_Mandate * ptRun = ptGlobals->ptMandateList;

              while (ptRun)
              {
                if (strcmp(ptRun->szMandate, ptCurrMandate->szMandate) == 0)
                {
                  ptLastNewestVersion		= ptRun->ptNewestVersion;
                  ptRun->ptNewestVersion	= ptCurrMandate;
                }
                ptRun = ptRun->ptNext;
              }
            } /* end-block sort mandate versions */

            /* append the mandate to the global list */
            *ptGlobals->pptMandateListEnd = ptCurrMandate;
            ptGlobals->pptMandateListEnd = & ptCurrMandate->ptNext;
          }
          break;


        case eDBType_TypeDateOfSign:
          if ( ! ptCurrMandate)
            iType = 0;
          else
            ptCurrMandate->lDateOfSign = tValue.lLong;
          break;
          
        case eDBType_TypeName:
          if ( ! ptCurrMandate)
            iType = 0;
          else
            ptCurrMandate->tDebitor.szName = tValue.szString;
          break;

        case eDBType_TypeIBAN:
          if ( ! ptCurrMandate)
            iType = 0;
          else
            ptCurrMandate->tDebitor.szIBAN = tValue.szString;
          break;

        case eDBType_TypeBIC:
          if ( ! ptCurrMandate)
            iType = 0;
          else
            ptCurrMandate->tDebitor.szBIC = tValue.szString;
          break;

        case eDBType_TypeValidSince:
          if ( ! ptCurrMandate)
            iType = 0;
          else
          {
            ptCurrMandate->iValidSincePersMsgID = tValue.iInt;

            // check if we had an undo 
            if (ptCurrMandate->iValidSincePersMsgID >= ptGlobals->iPersistentMsgID)
            {
            
               struct t_Mandate * ptRun = ptGlobals->ptMandateList;

              if (ptLastNewMandate == NULL)
              {
                // first element is the current element
                ptGlobals->ptMandateList = NULL;
                ptGlobals->pptMandateListEnd = & ptGlobals->ptMandateList;
              }
              else
              {
                while (ptRun)
                {
                  if (ptRun->ptNewestVersion == ptCurrMandate)
                    ptRun->ptNewestVersion = ptLastNewestVersion;

                  if (ptRun->ptNext == ptCurrMandate)
                    ptRun->ptNext = NULL;

                  ptRun = ptRun->ptNext;
                }
              }

              /* append the mandate to the global list */
              *ptGlobals->pptMandateListEnd = ptLastNewMandate;
              ptGlobals->pptMandateListEnd = & ptLastNewMandate->ptNext;

              free(ptCurrMandate->szMandate);
              free(ptCurrMandate->tDebitor.szName);
              free(ptCurrMandate->tDebitor.szIBAN);
              free(ptCurrMandate->tDebitor.szBIC);
              free(ptCurrMandate); 

            }
            else
              ptLastNewMandate = ptCurrMandate;

            ptCurrMandate = NULL;
          }
          break;
      }

      if (iType == 0)
        break;
        
        
    }

    fclose(ptDataBaseFile);

    if (lBytesToRead || ptCurrMandate)
      return (0);

  }
  else
  {
    fprintf(stderr, "warning: creating new database %s\n", szFileName);

    srand((unsigned int)time(NULL));
    int iUniquID = (rand() % 1000) * 1000;

    ptGlobals->iPersistentMsgID = iUniquID;
  }

  return (1);
}    



/****************************************************************************************
 *
 *
 * return
 *   bool		<success>
 ****************************************************************************************/
int _WriteDatabase(struct t_Globals * ptGlobals, char * szFileName)
{

  FILE * ptDataBaseFile = fopen(szFileName, "w");
  if (ptDataBaseFile)
  {
    struct t_Mandate *				ptCurrMandate	= ptGlobals->ptMandateList;
    long					lLenItem;
    union { int iInt; unsigned long lLong; char * szString; }
						tValue;

    tValue.iInt		= eDBType_TypePersistentMsgID;
    lLenItem		= sizeof(ptGlobals->iPersistentMsgID);
    tValue.iInt 	+= lLenItem;
    fwrite( & tValue.iInt, sizeof(tValue.iInt), 1, ptDataBaseFile);
    fwrite( & ptGlobals->iPersistentMsgID, lLenItem, 1, ptDataBaseFile);

    while (ptCurrMandate)
    {
      tValue.iInt	= eDBType_TypeMandateName;
      lLenItem		= strlen(ptCurrMandate->szMandate) + 1;
      tValue.iInt 	+= lLenItem;
      fwrite( & tValue.iInt, sizeof(tValue.iInt), 1, ptDataBaseFile);
      fwrite(ptCurrMandate->szMandate, lLenItem, 1, ptDataBaseFile);

      tValue.iInt	= eDBType_TypeDateOfSign;
      lLenItem		= sizeof(ptCurrMandate->lDateOfSign);
      tValue.iInt 	+= lLenItem;
      fwrite( & tValue.iInt, sizeof(tValue.iInt), 1, ptDataBaseFile);
      fwrite( & ptCurrMandate->lDateOfSign, lLenItem, 1, ptDataBaseFile);

      tValue.iInt	= eDBType_TypeName;
      lLenItem		= strlen(ptCurrMandate->tDebitor.szName) + 1;
      tValue.iInt 	+= lLenItem;
      fwrite( & tValue.iInt, sizeof(tValue.iInt), 1, ptDataBaseFile);
      fwrite(ptCurrMandate->tDebitor.szName, lLenItem, 1, ptDataBaseFile);

      tValue.iInt	= eDBType_TypeIBAN;
      lLenItem		= strlen(ptCurrMandate->tDebitor.szIBAN) + 1;
      tValue.iInt 	+= lLenItem;
      fwrite( & tValue.iInt, sizeof(tValue.iInt), 1, ptDataBaseFile);
      fwrite(ptCurrMandate->tDebitor.szIBAN, lLenItem, 1, ptDataBaseFile);

      tValue.iInt	= eDBType_TypeBIC;
      lLenItem		= strlen(ptCurrMandate->tDebitor.szBIC) + 1;
      tValue.iInt 	+= lLenItem;
      fwrite( & tValue.iInt, sizeof(tValue.iInt), 1, ptDataBaseFile);
      fwrite(ptCurrMandate->tDebitor.szBIC, lLenItem, 1, ptDataBaseFile);

      tValue.iInt	= eDBType_TypeValidSince;
      lLenItem		= sizeof(ptCurrMandate->iValidSincePersMsgID);
      tValue.iInt 	+= lLenItem;
      fwrite( & tValue.iInt, sizeof(tValue.iInt), 1, ptDataBaseFile);
      fwrite( & ptCurrMandate->iValidSincePersMsgID, lLenItem, 1, ptDataBaseFile);

      ptCurrMandate = ptCurrMandate->ptNext;
    }

    fclose(ptDataBaseFile);

    if (ptCurrMandate)
      return (0);

  }


  return (1);
}    


/****************************************************************************************
 *
 *
 * return
 *   bool		<success>
 ****************************************************************************************/
int _GiveUpDatabase(struct t_Globals * ptGlobals)
{
  struct t_Mandate *				ptCurrMandate	= ptGlobals->ptMandateList;

  while (ptCurrMandate)
  {
    free(ptCurrMandate->szMandate);
    free(ptCurrMandate->tDebitor.szName);
    free(ptCurrMandate->tDebitor.szIBAN);
    free(ptCurrMandate->tDebitor.szBIC);

    struct t_Mandate * ptToDelete = ptCurrMandate;
    ptCurrMandate = ptCurrMandate->ptNext;

    free (ptToDelete);
  }

  return (1);
}

