#include "config.h"

#ifdef MEM_TEST

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include "memory.h"

#undef malloc
#undef free
#undef realloc
#undef main


/***********************************************************************************
 ***********************************************************************************/
struct t_Portion
{
  struct t_Portion *	ptNext;
  struct t_Portion *	ptPrevious;
  void *		vptUsedPtr;
  int			iLineNr;
  size_t		iAllocatedLength;
  char			szFileName[sizeof (void *)];	/* length of back pointer to this structure */
};

/***********************************************************************************
 ***********************************************************************************/
static struct t_Portion *	ptFirstPortion	= NULL;
static struct t_Portion *	ptLastPortion	= NULL;


/***********************************************************************************
 ***********************************************************************************/
static void _memtest(void);


/***********************************************************************************
 ***********************************************************************************/
int main (int argc, char **argv)
{
  int iReturn = _mt_main(argc, argv);

  _memtest();

  return (iReturn);
}


/***********************************************************************************
 ***********************************************************************************/
void *_mt_malloc(char * szFileName, int iLineNr, size_t size)
{
  size_t		iPortionLength	= sizeof(struct t_Portion) + strlen(szFileName);
  size_t		iAlignMask	= (1 << (sizeof (void *))) - 1;
  size_t		iEntireLength;
  struct t_Portion *	ptReturn;

  iPortionLength = (iPortionLength & iAlignMask) ? (iPortionLength + iAlignMask) & ( ~ iAlignMask) : iPortionLength;
  iEntireLength = size + iPortionLength;

  ptReturn = (struct t_Portion *) malloc (iEntireLength);


  ptReturn->ptNext		= NULL;
  ptReturn->ptPrevious		= ptLastPortion;
  ptReturn->vptUsedPtr		= ((char *) ptReturn) + iPortionLength;
  ptReturn->iLineNr		= iLineNr;
  ptReturn->iAllocatedLength	= size;

  strcpy(ptReturn->szFileName, szFileName);

  *((struct t_Portion **) (((char *) ptReturn->vptUsedPtr) - sizeof(void *))) = ptReturn;


  if ( ! ptFirstPortion)
    ptFirstPortion = ptReturn;
  
  if (ptReturn->ptPrevious)
    ptReturn->ptPrevious->ptNext = ptReturn;

  ptLastPortion = ptReturn;

  return (ptReturn->vptUsedPtr);
  
}



/***********************************************************************************
 ***********************************************************************************/
void _mt_free(char * szFilename, int iLineNr, void *ptr)
{
  struct t_Portion **	pptRun	= & ptFirstPortion;
  struct t_Portion *    ptLast	= NULL;
  

  while (*pptRun && ((*pptRun)->vptUsedPtr != ptr))
  {
    ptLast = *pptRun;
    pptRun =  & (*pptRun)->ptNext;
  }

  if (*pptRun)
  {
    struct t_Portion * ptToDelete = *pptRun;
    
    *pptRun = (*pptRun)->ptNext;

    if (*pptRun)
    {
      if ((*pptRun)->ptPrevious)
        (*pptRun)->ptPrevious = ptLast;
    }
    else
      ptLastPortion = ptLast;
    
    free (ptToDelete);

  }
  else
    fprintf(stderr, "mem_error: %s: % 4d: invalid free ptr %p\n", szFilename, iLineNr, ptr);
  
}

/***********************************************************************************
 ***********************************************************************************/
void *_mt_realloc(char * szFilename, int iLineNr, void *ptr, size_t size)
{


  if (ptr)
  {
    struct t_Portion *	ptThis	= ptFirstPortion;


    while (ptThis && (ptThis->vptUsedPtr != ptr))
      ptThis = ptThis->ptNext;

    int	iPortionLength	= ((char *) (ptThis->vptUsedPtr)) - ((char *) ptThis);
    int	iEntireLength	= size + iPortionLength + sizeof(void *); 

    ptThis = (struct t_Portion *) realloc (ptThis, iEntireLength);

    ptThis->vptUsedPtr		= ((char *) ptThis) + iPortionLength;
    ptThis->iAllocatedLength	= size;

    *((struct t_Portion **) (((char *) ptThis->vptUsedPtr) - sizeof(void *))) = ptThis;

    return (ptThis->vptUsedPtr);
  }
  else
    return (_mt_malloc(szFilename, iLineNr, size));
}


/***********************************************************************************
 ***********************************************************************************/
void _memtest(void)
{
  struct t_Portion *	ptRun	= ptFirstPortion;
  char			ptShowBuffer[18];

  ptShowBuffer[8] = ' ';
  ptShowBuffer[17] = '\0';
  
  while (ptRun)
  {
    long	iI;
    int		iPosCounter = 0;

    fprintf(stderr, "memtest: %s: % 5d: length %-6d: still allocated\n", ptRun->szFileName, ptRun->iLineNr, ptRun->iAllocatedLength);

    for (iI = 0; iI < ptRun->iAllocatedLength; iI++)
    {
      int iCurrChar = *(((unsigned char *) (ptRun->vptUsedPtr)) + iI);
      if (iPosCounter == 0)
        fprintf(stderr, "%08lx  ", iI);

      fprintf(stderr, "%02x ", iCurrChar);

      if (!isprint(iCurrChar))
        iCurrChar = '.';
      ptShowBuffer[iPosCounter] = iCurrChar;

      /* check for end of line */
      if (iPosCounter == 16)
      {
        iPosCounter = 0;
        fprintf(stderr, "  %s\n", ptShowBuffer);
      }
      else
      {
        /* separator position ? */
        if (iPosCounter == 7)
        {
          fprintf(stderr, "  ");
          iPosCounter++;
        }

        iPosCounter++;
      }
    }

    /* handle final row output */
    if (iPosCounter)
    {
      /* fill up everything with spaces */
      while (iPosCounter < 17)
      {
        if (iPosCounter == 7)
          fprintf(stderr, "  ");
        else
          fprintf(stderr, "   ");
        ptShowBuffer[iPosCounter++] = ' ';
      }
      /* and give out the preview buffer */
      fprintf(stderr, "  %s\n", ptShowBuffer);
    }
    fprintf(stderr, "\n");

    ptRun = ptRun->ptNext;
  }  
}


/***********************************************************************************
 ***********************************************************************************/
void memtest_r(void)
{
  struct t_Portion *	ptRun	= ptLastPortion;

  while (ptRun)
  {
    fprintf(stderr, "memtest: %s: % 5d: length %d: still allocated\n", ptRun->szFileName, ptRun->iLineNr, ptRun->iAllocatedLength);
    ptRun = ptRun->ptPrevious;
  }  
}



#endif /* MEM_TEST */
