#include "config.h"

#include <time.h>
#include <string.h>


#include "sepa.h"
#include "sepadata.h"


#include "memory.h"

/****************************************************************************************
 ****************************************************************************************/
#define SPC_LEN 80

static char szSpaces[SPC_LEN + 1];



static int _iCreditTransferTransaction(struct t_Globals * ptGlobals, struct t_CdtTrfTx * ptPmt);
static int _iCdtPmtInf(struct t_Globals * ptGlobals);
static int _iGrpHdr(struct t_Globals * ptGlobals);
static int _iCstmrCdtTrfInitn(struct t_Globals * ptGlobals);
static int _iPain00100203 (struct t_Globals * ptGlobals);
static int _iPain00800202 (struct t_Globals * ptGlobals);
static int _iCstmrDrctDbtInitn(struct t_Globals * ptGlobals);
static int _iDbtPmtInfLst(struct t_Globals * ptGlobals);
static int _iDbtPmtInf(struct t_Globals * ptGlobals, struct t_DbtTrfTx * ptTransferList);
static int _iDirectDebitTransferTransaction(struct t_Globals * ptGlobals, struct t_DbtTrfTx * ptTransaction);
static int _iAmdmntInfDtls(struct t_Globals * ptGlobals, struct t_DbtTrfTx * ptTransaction);
static int _iPrintHeader(struct t_Globals * ptGlobals);
static int _iIndent(struct t_Globals * ptGlobals, int iIn);


/****************************************************************************************
 *
 *
 *
 ****************************************************************************************/
int iDoCreditTransfer(struct t_Globals * ptGlobals)
{
  _iPrintHeader(ptGlobals);
  return (_iPain00100203(ptGlobals));
}


/****************************************************************************************
 *
 *
 *
 ****************************************************************************************/
int iDoDirectDebitInitiation(struct t_Globals * ptGlobals)
{
  _iPrintHeader(ptGlobals);
  return (_iPain00800202(ptGlobals));
}


/****************************************************************************************
 *
 *
 *
 ****************************************************************************************/
struct t_Mandate * FindMandate(struct t_Globals * ptGlobals, char * szMandate)
{
  struct t_Mandate * ptRun = ptGlobals->ptMandateList;

  while (ptRun)
  {
    if (strcmp(ptRun->szMandate, szMandate) == 0)
    {
      return (ptRun);
    }
    ptRun = ptRun->ptNext;
  }

  return (NULL);
}

/****************************************************************************************
 *
 *
 *
 ****************************************************************************************/
int _iIndent(struct t_Globals * ptGlobals, int iIn)
{
  
  if (iIn)
  {
    ptGlobals->szCurrIndent -= 2;
    if (ptGlobals->szCurrIndent < szSpaces)
      ptGlobals->szCurrIndent = szSpaces;
  }
  else
  {
    ptGlobals->szCurrIndent += 2;
    if (ptGlobals->szCurrIndent > szSpaces + SPC_LEN)
      ptGlobals->szCurrIndent = szSpaces + SPC_LEN;
  }

  return (1);

}

/****************************************************************************************
 *
 *
 *
 ****************************************************************************************/
int _iPrintHeader(struct t_Globals * ptGlobals)
{

  int iI;

  for (iI = 0; iI < SPC_LEN; iI++)
    szSpaces[iI] = ' ';
  szSpaces[iI] = '\0';
  ptGlobals->szCurrIndent = szSpaces + SPC_LEN;

  fprintf (ptGlobals->ptOutFile, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");

  return (1);
}


/****************************************************************************************
 *
 *
 *
 ****************************************************************************************/
int _iPain00100203 (struct t_Globals * ptGlobals)
{
  fprintf (ptGlobals->ptOutFile, "%s<Document xmlns=\"urn:iso:std:iso:20022:tech:xsd:pain.001.002.03\"\n          xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n          xsi:schemaLocation=\"urn:iso:std:iso:20022:tech:xsd:pain.001.002.03\n          pain.001.002.03.xsd\">\n",
           ptGlobals->szCurrIndent);
  _iIndent(ptGlobals, 1);
  _iCstmrCdtTrfInitn(ptGlobals);
  _iIndent(ptGlobals, 0);
  fprintf (ptGlobals->ptOutFile, "%s</Document>\n", ptGlobals->szCurrIndent);

  return (1);
}


/****************************************************************************************
 *
 *
 *
 ****************************************************************************************/
int _iCstmrCdtTrfInitn(struct t_Globals * ptGlobals)
{
  fprintf (ptGlobals->ptOutFile, "%s<CstmrCdtTrfInitn>\n", ptGlobals->szCurrIndent);
  _iIndent(ptGlobals, 1);
  _iGrpHdr(ptGlobals);
  /* just one Payment Instruction Information in this implementation */
  _iCdtPmtInf(ptGlobals);
  _iIndent(ptGlobals, 0);
  fprintf (ptGlobals->ptOutFile, "%s</CstmrCdtTrfInitn>\n", ptGlobals->szCurrIndent);

  return (1);
}


/****************************************************************************************
 *
 *
 *
 ****************************************************************************************/
int _iGrpHdr(struct t_Globals * ptGlobals)
{
  fprintf (ptGlobals->ptOutFile, "%s<GrpHdr>\n", ptGlobals->szCurrIndent);
  _iIndent(ptGlobals, 1);
  
  fprintf (ptGlobals->ptOutFile, "%s<MsgId>MID-%08x</MsgId>\n", ptGlobals->szCurrIndent,
                                 ptGlobals->iPersistentMsgID);
  fprintf (ptGlobals->ptOutFile, "%s<CreDtTm>%s</CreDtTm>\n", ptGlobals->szCurrIndent,
                                 GetISODateTimeString(time(NULL)));
  fprintf (ptGlobals->ptOutFile, "%s<NbOfTxs>%d</NbOfTxs>\n", ptGlobals->szCurrIndent,
                                 ptGlobals->iPmtCount);
  fprintf (ptGlobals->ptOutFile, "%s<InitgPty>\n", ptGlobals->szCurrIndent);
  _iIndent(ptGlobals, 1);
  fprintf (ptGlobals->ptOutFile, "%s<Nm>%s</Nm>\n", ptGlobals->szCurrIndent,
                                 ptGlobals->ptInitiator->szName);
  _iIndent(ptGlobals, 0);
  fprintf (ptGlobals->ptOutFile, "%s</InitgPty>\n", ptGlobals->szCurrIndent);
  _iIndent(ptGlobals, 0);
  fprintf (ptGlobals->ptOutFile, "%s</GrpHdr>\n", ptGlobals->szCurrIndent);

  return (1);
}


/****************************************************************************************
 *
 *
 *
 ****************************************************************************************/
int _iCdtPmtInf(struct t_Globals * ptGlobals)
{

  fprintf (ptGlobals->ptOutFile, "%s<PmtInf>\n", ptGlobals->szCurrIndent);
  _iIndent(ptGlobals, 1);

  fprintf (ptGlobals->ptOutFile, "%s<PmtInfId>PIID-%08x</PmtInfId>\n", ptGlobals->szCurrIndent,
                                 ptGlobals->iPersistentMsgID);
  fprintf (ptGlobals->ptOutFile, "%s<PmtMtd>TRF</PmtMtd>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s<BtchBookg>%s</BtchBookg>\n", ptGlobals->szCurrIndent,
                                 GetTrueFalseString(ptGlobals->iBtchBookg));
  fprintf (ptGlobals->ptOutFile, "%s<NbOfTxs>%d</NbOfTxs>\n", ptGlobals->szCurrIndent,
                                 ptGlobals->iPmtCount);
  fprintf (ptGlobals->ptOutFile, "%s<CtrlSum>%s</CtrlSum>\n", ptGlobals->szCurrIndent,
                                 GetValueFromCentString(ptGlobals->iPmtSumInCent));

  fprintf (ptGlobals->ptOutFile, "%s<PmtTpInf>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  <SvcLvl>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s    <Cd>SEPA</Cd>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  </SvcLvl>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s</PmtTpInf>\n", ptGlobals->szCurrIndent);

  fprintf (ptGlobals->ptOutFile, "%s<ReqdExctnDt>%s</ReqdExctnDt>\n", ptGlobals->szCurrIndent,
                                  GetISODateString(ptGlobals->iExecTimeStamp));

  fprintf (ptGlobals->ptOutFile, "%s<Dbtr>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  <Nm>%s</Nm>\n", ptGlobals->szCurrIndent,
                                      ptGlobals->ptDebitor->szName);
  fprintf (ptGlobals->ptOutFile, "%s</Dbtr>\n", ptGlobals->szCurrIndent);

  
  fprintf (ptGlobals->ptOutFile, "%s<DbtrAcct>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  <Id>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s    <IBAN>%s</IBAN>\n", ptGlobals->szCurrIndent,
                                        ptGlobals->ptDebitor->szIBAN);
  fprintf (ptGlobals->ptOutFile, "%s  </Id>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s</DbtrAcct>\n", ptGlobals->szCurrIndent);

  fprintf (ptGlobals->ptOutFile, "%s<DbtrAgt>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  <FinInstnId>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s    <BIC>%s</BIC>\n", ptGlobals->szCurrIndent,
                                        ptGlobals->ptDebitor->szBIC);
  fprintf (ptGlobals->ptOutFile, "%s  </FinInstnId>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s</DbtrAgt>\n", ptGlobals->szCurrIndent);

  fprintf (ptGlobals->ptOutFile, "%s<ChrgBr>SLEV</ChrgBr>\n", ptGlobals->szCurrIndent);


  struct t_CdtTrfTx * ptPmt = ptGlobals->ptTxList.ptCdtTrfTxList;
  while (ptPmt)
  {
    _iCreditTransferTransaction(ptGlobals, ptPmt);

    ptPmt = ptPmt->ptNext;
  }
  
  _iIndent(ptGlobals, 0);
  fprintf (ptGlobals->ptOutFile, "%s</PmtInf>\n", ptGlobals->szCurrIndent);

  return (1);
}

/****************************************************************************************
 *
 *
 *
 ****************************************************************************************/
int _iCreditTransferTransaction(struct t_Globals * ptGlobals, struct t_CdtTrfTx * ptPmt)
{
  fprintf (ptGlobals->ptOutFile, "%s<CdtTrfTxInf>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  <PmtId>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s    <EndToEndId>%s</EndToEndId>\n", ptGlobals->szCurrIndent,
                                        "NOTPROVIDED");
  fprintf (ptGlobals->ptOutFile, "%s  </PmtId>\n", ptGlobals->szCurrIndent);

  fprintf (ptGlobals->ptOutFile, "%s  <Amt>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s    <InstdAmt Ccy=\"EUR\">%s</InstdAmt>\n", ptGlobals->szCurrIndent,
                                        GetValueFromCentString(ptPmt->iValueInCent));
  fprintf (ptGlobals->ptOutFile, "%s  </Amt>\n", ptGlobals->szCurrIndent);

  fprintf (ptGlobals->ptOutFile, "%s  <CdtrAgt>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s    <FinInstnId>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s      <BIC>%s</BIC>\n", ptGlobals->szCurrIndent,
                                          ptPmt->ptCreditor->szBIC);
  fprintf (ptGlobals->ptOutFile, "%s    </FinInstnId>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  </CdtrAgt>\n", ptGlobals->szCurrIndent);

  fprintf (ptGlobals->ptOutFile, "%s  <Cdtr>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s    <Nm>%s</Nm>\n", ptGlobals->szCurrIndent,
                                         ptPmt->ptCreditor->szName);
  fprintf (ptGlobals->ptOutFile, "%s  </Cdtr>\n", ptGlobals->szCurrIndent);

  fprintf (ptGlobals->ptOutFile, "%s  <CdtrAcct>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s    <Id>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s      <IBAN>%s</IBAN>\n", ptGlobals->szCurrIndent,
                                          ptPmt->ptCreditor->szIBAN);
  fprintf (ptGlobals->ptOutFile, "%s    </Id>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  </CdtrAcct>\n", ptGlobals->szCurrIndent);


  fprintf (ptGlobals->ptOutFile, "%s  <RmtInf>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s    <Ustrd>%s</Ustrd>\n", ptGlobals->szCurrIndent,
                                        ptPmt->szPurpose);
  fprintf (ptGlobals->ptOutFile, "%s  </RmtInf>\n", ptGlobals->szCurrIndent);

  fprintf (ptGlobals->ptOutFile, "%s</CdtTrfTxInf>\n", ptGlobals->szCurrIndent);

  return (1);
}



/****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************/

/****************************************************************************************
 *
 *
 *
 ****************************************************************************************/
int _iPain00800202 (struct t_Globals * ptGlobals)
{
  int			iReturn;

  fprintf (ptGlobals->ptOutFile, "%s<Document xmlns=\"urn:iso:std:iso:20022:tech:xsd:pain.008.002.02\"\n          xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n          xsi:schemaLocation=\"urn:iso:std:iso:20022:tech:xsd:pain.008.002.02\n          pain.008.002.02.xsd\">\n",
           ptGlobals->szCurrIndent);
  _iIndent(ptGlobals, 1);
  iReturn = _iCstmrDrctDbtInitn(ptGlobals);
  _iIndent(ptGlobals, 0);
  fprintf (ptGlobals->ptOutFile, "%s</Document>\n", ptGlobals->szCurrIndent);

  return (iReturn);
}

/****************************************************************************************
 *
 *
 *
 ****************************************************************************************/
int _iCstmrDrctDbtInitn(struct t_Globals * ptGlobals)
{
  int			iReturn;

  fprintf (ptGlobals->ptOutFile, "%s<CstmrDrctDbtInitn>\n", ptGlobals->szCurrIndent);
  _iIndent(ptGlobals, 1);

  if ((iReturn = _iGrpHdr(ptGlobals)))
    iReturn = _iDbtPmtInfLst(ptGlobals);

  _iIndent(ptGlobals, 0);
  fprintf (ptGlobals->ptOutFile, "%s</CstmrDrctDbtInitn>\n", ptGlobals->szCurrIndent);

  return (iReturn);
}

/****************************************************************************************
 *
 *
 *
 ****************************************************************************************/
int _iDbtPmtInfLst(struct t_Globals * ptGlobals)
{
  struct t_DbtTrfTx *   ptTransferList = ptGlobals->ptTxList.ptDrctDbtTrfTxList;
  int			iReturn = 1;

  while (ptTransferList && iReturn)
  {
    iReturn = _iDbtPmtInf(ptGlobals, ptTransferList);
    ptTransferList = ptTransferList->ptNextSubType;
  }

  return (iReturn);
}


/****************************************************************************************
 *
 *
 *
 ****************************************************************************************/
int _iDbtPmtInf(struct t_Globals * ptGlobals, struct t_DbtTrfTx * ptTransferList)
{
  char * szLocalInstrument = (ptGlobals->iAction == eAction_DirectDebitCore) ? "CORE" : 
                             (ptGlobals->iAction == eAction_DirectDebitB2B)  ? "B2B" : "";
  char * szSequence        = (ptTransferList->iType == eTASubType_DirectDebitFirst)  ? "FRST" :
                             (ptTransferList->iType == eTASubType_DirectDebitOneOff) ? "OOFF" : 
                             (ptTransferList->iType == eTASubType_DirectDebitRepeat) ? "RCUR" : 
                             (ptTransferList->iType == eTASubType_DirectDebitFinal)   ? "FNAL" : ""; 

  fprintf (ptGlobals->ptOutFile, "%s<PmtInf>\n", ptGlobals->szCurrIndent);
  _iIndent(ptGlobals, 1);

  fprintf (ptGlobals->ptOutFile, "%s<PmtInfId>PIID-%08x</PmtInfId>\n", ptGlobals->szCurrIndent,
                                 ptGlobals->iPersistentMsgID);
  fprintf (ptGlobals->ptOutFile, "%s<PmtMtd>DD</PmtMtd>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s<BtchBookg>%s</BtchBookg>\n", ptGlobals->szCurrIndent,
                                 GetTrueFalseString(ptGlobals->iBtchBookg));
  fprintf (ptGlobals->ptOutFile, "%s<NbOfTxs>%d</NbOfTxs>\n", ptGlobals->szCurrIndent,
                                 ptTransferList->iSTCount);
  fprintf (ptGlobals->ptOutFile, "%s<CtrlSum>%s</CtrlSum>\n", ptGlobals->szCurrIndent,
                                 GetValueFromCentString(ptTransferList->iSTValueInCent));

  fprintf (ptGlobals->ptOutFile, "%s<PmtTpInf>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  <SvcLvl>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s    <Cd>SEPA</Cd>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  </SvcLvl>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  <LclInstrm>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s    <Cd>%s</Cd>\n", ptGlobals->szCurrIndent, szLocalInstrument);
  fprintf (ptGlobals->ptOutFile, "%s  </LclInstrm>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  <SeqTp>%s</SeqTp>\n", ptGlobals->szCurrIndent, szSequence);
  fprintf (ptGlobals->ptOutFile, "%s</PmtTpInf>\n", ptGlobals->szCurrIndent);


  fprintf (ptGlobals->ptOutFile, "%s<ReqdColltnDt>%s</ReqdColltnDt>\n", ptGlobals->szCurrIndent,
                                    GetISODateString(ptGlobals->iExecTimeStamp));

  fprintf (ptGlobals->ptOutFile, "%s<Cdtr>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  <Nm>%s</Nm>\n", ptGlobals->szCurrIndent,
                                      ptGlobals->ptCreditor->szName);
  fprintf (ptGlobals->ptOutFile, "%s</Cdtr>\n", ptGlobals->szCurrIndent);

  fprintf (ptGlobals->ptOutFile, "%s<CdtrAcct>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  <Id>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s    <IBAN>%s</IBAN>\n", ptGlobals->szCurrIndent,
                                        ptGlobals->ptCreditor->szIBAN);
  fprintf (ptGlobals->ptOutFile, "%s  </Id>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s</CdtrAcct>\n", ptGlobals->szCurrIndent);

  fprintf (ptGlobals->ptOutFile, "%s<CdtrAgt>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  <FinInstnId>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s    <BIC>%s</BIC>\n", ptGlobals->szCurrIndent,
                                        ptGlobals->ptCreditor->szBIC);
  fprintf (ptGlobals->ptOutFile, "%s  </FinInstnId>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s</CdtrAgt>\n", ptGlobals->szCurrIndent);

  fprintf (ptGlobals->ptOutFile, "%s<ChrgBr>SLEV</ChrgBr>\n", ptGlobals->szCurrIndent);

  fprintf (ptGlobals->ptOutFile, "%s<CdtrSchmeId>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  <Id>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s    <PrvtId>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s      <Othr>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s        <Id>%s</Id>\n", ptGlobals->szCurrIndent,
                                            ptGlobals->szIdentification);
  fprintf (ptGlobals->ptOutFile, "%s        <SchmeNm>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s          <Prtry>SEPA</Prtry>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s        </SchmeNm>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s      </Othr>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s    </PrvtId>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  </Id>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s</CdtrSchmeId>\n", ptGlobals->szCurrIndent);

  while (ptTransferList)
  {
    _iDirectDebitTransferTransaction(ptGlobals, ptTransferList);
    ptTransferList = ptTransferList->ptNext;
  }

  _iIndent(ptGlobals, 0);
  fprintf (ptGlobals->ptOutFile, "%s</PmtInf>\n", ptGlobals->szCurrIndent);

  return (1);
}

/****************************************************************************************
 *
 *
 *
 ****************************************************************************************/
int _iDirectDebitTransferTransaction(struct t_Globals * ptGlobals, struct t_DbtTrfTx * ptTransaction)
{

  fprintf (ptGlobals->ptOutFile, "%s<DrctDbtTxInf>\n", ptGlobals->szCurrIndent);
  _iIndent(ptGlobals, 1);


  /************************************************** */
  /* <PmtId> */
  /************************************************** */
  fprintf (ptGlobals->ptOutFile, "%s<PmtId>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  <EndToEndId>%s</EndToEndId>\n", ptGlobals->szCurrIndent,
                                      "NOTPROVIDED");
  fprintf (ptGlobals->ptOutFile, "%s</PmtId>\n", ptGlobals->szCurrIndent);

  /************************************************** */
  /* <InstdAmt> */
  /************************************************** */
  fprintf (ptGlobals->ptOutFile, "%s<InstdAmt Ccy=\"EUR\">%s</InstdAmt>\n", ptGlobals->szCurrIndent,
                                    GetValueFromCentString(ptTransaction->iValueInCent));
                                    
  /************************************************** */
  /* <DrctDbtTx> */
  /************************************************** */
  fprintf (ptGlobals->ptOutFile, "%s<DrctDbtTx>\n", ptGlobals->szCurrIndent);
  _iIndent(ptGlobals, 1);
  fprintf (ptGlobals->ptOutFile, "%s<MndtRltdInf>\n", ptGlobals->szCurrIndent);
  _iIndent(ptGlobals, 1);
  fprintf (ptGlobals->ptOutFile, "%s<MndtId>%s</MndtId>\n", ptGlobals->szCurrIndent,
                                    ptTransaction->ptMandate->ptNewestVersion->szMandate);
  fprintf (ptGlobals->ptOutFile, "%s<DtOfSgntr>%s</DtOfSgntr>\n", ptGlobals->szCurrIndent,
                                    GetISODateString(ptTransaction->ptMandate->ptNewestVersion->lDateOfSign));
  fprintf (ptGlobals->ptOutFile, "%s<AmdmntInd>%s</AmdmntInd>\n", ptGlobals->szCurrIndent,
                                    GetTrueFalseString(ptTransaction->ptMandate->ptNewestVersion->iMndtChanged != eMndtChange_None));
  if (    (ptTransaction->ptMandate->ptNewestVersion->iMndtChanged == eMndtChange_IBAN)
       || (ptTransaction->ptMandate->ptNewestVersion->iMndtChanged == eMndtChange_BIC))
  {
    _iAmdmntInfDtls(ptGlobals, ptTransaction);
  }
  _iIndent(ptGlobals, 0);
  fprintf (ptGlobals->ptOutFile, "%s</MndtRltdInf>\n", ptGlobals->szCurrIndent);
  _iIndent(ptGlobals, 0);
  fprintf (ptGlobals->ptOutFile, "%s</DrctDbtTx>\n", ptGlobals->szCurrIndent);



  /************************************************** */
  /* <DbtrAgt>, <Dbtr>, <DbtrAcct> */
  /************************************************** */

  fprintf (ptGlobals->ptOutFile, "%s<DbtrAgt>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  <FinInstnId>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s    <BIC>%s</BIC>\n", ptGlobals->szCurrIndent,
                                        ptTransaction->ptMandate->ptNewestVersion->tDebitor.szBIC);
  fprintf (ptGlobals->ptOutFile, "%s  </FinInstnId>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s</DbtrAgt>\n", ptGlobals->szCurrIndent);

  fprintf (ptGlobals->ptOutFile, "%s<Dbtr>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  <Nm>%s</Nm>\n", ptGlobals->szCurrIndent,
                                        ptTransaction->ptMandate->ptNewestVersion->tDebitor.szName);
  fprintf (ptGlobals->ptOutFile, "%s</Dbtr>\n", ptGlobals->szCurrIndent);

  fprintf (ptGlobals->ptOutFile, "%s<DbtrAcct>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  <Id>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s    <IBAN>%s</IBAN>\n", ptGlobals->szCurrIndent,
                                        ptTransaction->ptMandate->ptNewestVersion->tDebitor.szIBAN);
  fprintf (ptGlobals->ptOutFile, "%s  </Id>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s</DbtrAcct>\n", ptGlobals->szCurrIndent);




  
  /************************************************** */
  /* <RmtInf> */
  /************************************************** */
  fprintf (ptGlobals->ptOutFile, "%s<RmtInf>\n", ptGlobals->szCurrIndent);
  fprintf (ptGlobals->ptOutFile, "%s  <Ustrd>%s</Ustrd>\n", ptGlobals->szCurrIndent,
                                      ptTransaction->szPurpose);
  fprintf (ptGlobals->ptOutFile, "%s</RmtInf>\n", ptGlobals->szCurrIndent);


  _iIndent(ptGlobals, 0);
  fprintf (ptGlobals->ptOutFile, "%s</DrctDbtTxInf>\n", ptGlobals->szCurrIndent);

  return (1);

}





/****************************************************************************************
 *
 *
 *
 ****************************************************************************************/
int _iAmdmntInfDtls(struct t_Globals * ptGlobals, struct t_DbtTrfTx * ptTransaction)
{
  fprintf (ptGlobals->ptOutFile, "%s<AmdmntInfDtls>\n", ptGlobals->szCurrIndent);
  _iIndent(ptGlobals, 1);

  if (ptTransaction->ptMandate->ptNewestVersion->iMndtChanged == eMndtChange_IBAN)
  {
    fprintf (ptGlobals->ptOutFile, "%s<OrgnlDbtrAcct>\n", ptGlobals->szCurrIndent);
    fprintf (ptGlobals->ptOutFile, "%s  <Id>\n", ptGlobals->szCurrIndent);
    fprintf (ptGlobals->ptOutFile, "%s    <IBAN>%s</IBAN>\n", ptGlobals->szCurrIndent,
                                          ptTransaction->ptMandate->ptLastUsedVersion->tDebitor.szIBAN);
    fprintf (ptGlobals->ptOutFile, "%s  </Id>\n", ptGlobals->szCurrIndent);
    fprintf (ptGlobals->ptOutFile, "%s</OrgnlDbtrAcct>\n", ptGlobals->szCurrIndent);
  }

  if (ptTransaction->ptMandate->ptNewestVersion->iMndtChanged == eMndtChange_BIC)
  {
    fprintf (ptGlobals->ptOutFile, "%s<OrgnlDbtrAgt>\n", ptGlobals->szCurrIndent);
    fprintf (ptGlobals->ptOutFile, "%s  <FinInstnId>\n", ptGlobals->szCurrIndent);
    fprintf (ptGlobals->ptOutFile, "%s    <Othr>\n", ptGlobals->szCurrIndent);
    fprintf (ptGlobals->ptOutFile, "%s      <Id>SMNDA</Id>\n", ptGlobals->szCurrIndent);
    fprintf (ptGlobals->ptOutFile, "%s    </Othr>\n", ptGlobals->szCurrIndent);
    fprintf (ptGlobals->ptOutFile, "%s  </FinInstnId>\n", ptGlobals->szCurrIndent);
    fprintf (ptGlobals->ptOutFile, "%s</OrgnlDbtrAgt>\n", ptGlobals->szCurrIndent);
  }

  _iIndent(ptGlobals, 0);
  fprintf (ptGlobals->ptOutFile, "%s</AmdmntInfDtls>\n", ptGlobals->szCurrIndent);

  return (1);
}

/****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************/

