
#ifndef SEPA_H
#define SEPA_H

#include <stdio.h>
#include <time.h>

enum eAction
{
  eAction_CreditTransfer,
  eAction_DirectDebitCore,
  eAction_DirectDebitB2B,
};

enum eTASubType
{
  eTASubType_InvalidType = 0x000,

  eTASubType_CreditTransferType = 0x100,
  eTASubType_DebitTransferType  = 0x200,

  eTASubType_Credit = eTASubType_CreditTransferType,
  eTASubType_Salary,

  eTASubType_DirectDebitFirst = eTASubType_DebitTransferType,
  eTASubType_DirectDebitRepeat,
  eTASubType_DirectDebitFinal,
  eTASubType_DirectDebitOneOff,

  eTASubType_DirectDebitEmitDefault

};

enum eMndtChange
{
  eMndtChange_None,
  eMndtChange_Name,
  eMndtChange_IBAN,
  eMndtChange_BIC,
};


/****************************************************************************************
 ****************************************************************************************/
struct t_Globals
{
  FILE * 		ptOutFile;
  char * 		szCurrIndent;

  enum eAction 		iAction;

  enum eTASubType	iWhatToEmit;
  
  int			iPersistentMsgID;
  int   		iBtchBookg;
  
  int   		iExecTimeStamp;

  struct t_CredDebt *	ptInitiator;
  struct t_CredDebt *	ptCreditor;
  struct t_CredDebt *	ptDebitor;
  char *		szIdentification;

  struct t_Mandate *	ptMandateList;
  struct t_Mandate **	pptMandateListEnd;

  
  union
  {
    struct t_CdtTrfTx * ptCdtTrfTxList;
    struct t_DbtTrfTx * ptDrctDbtTrfTxList;
  } ptTxList;


  struct t_CdtTrfTx *	ptLICdtTrfTx;
  struct t_DbtTrfTx *	ptLIDrctDbtTrfTx;

  int 			iRecordCount;
  int 			iPmtCount;
  int 			iPmtSumInCent;

  
};


/****************************************************************************************
 ****************************************************************************************/
struct t_CdtTrfTx
{
  enum eTASubType	iType;
  struct t_CredDebt *	ptCreditor;
  int			iValueInCent;
  char *		szPurpose;
  
  struct t_CdtTrfTx *	ptNext;
};


/****************************************************************************************
 ****************************************************************************************/
struct t_DbtTrfTx
{
  enum eTASubType	iType;
  struct t_Mandate *	ptMandate;
  int			iValueInCent;
  int			iSTCount;
  int			iSTValueInCent;
  char *		szPurpose;
  
  struct t_DbtTrfTx *	ptNext;
  struct t_DbtTrfTx *	ptNextSubType;
};


/****************************************************************************************
 ****************************************************************************************/
struct t_CredDebt
{
  char * 		szName;
  char * 		szIBAN;
  char * 		szBIC;
};

/****************************************************************************************
 ****************************************************************************************/
struct t_Mandate
{
  char * 		szMandate;
  time_t		lDateOfSign;
  enum eMndtChange	iMndtChanged;
  struct t_CredDebt	tDebitor;

  int			iValidSincePersMsgID;
  struct t_Mandate *	ptNext;
  struct t_Mandate *	ptLastUsedVersion;
  struct t_Mandate *	ptNewestVersion;
};




/****************************************************************************************
 ****************************************************************************************/
extern int iDoCreditTransfer(struct t_Globals * ptGlobals);
extern int iDoDirectDebitInitiation(struct t_Globals * ptGlobals);
extern struct t_Mandate * FindMandate(struct t_Globals * ptGlobals, char * szMandate);

#endif
